# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module,invalid-name
"""
Tests for the migrations of the attributes, extras and settings from EAV to JSONB
Migration 0037_attributes_extras_settings_json
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy

import six
from six.moves import range

from django.db import transaction
from django.utils.encoding import python_2_unicode_compatible

from aiida.backends.djsite.db.subtests.migrations.test_migrations_common import TestMigrations

# The following sample dictionary can be used for the conversion test of attributes and extras
SAMPLE_DICT = {
    'bool': True,
    '001': 2,
    '17': 'string',
    'integer': 12,
    'float': 26.2,
    'string': 'a string',
    'dict': {
        '25': [True, False],
        'a': 'b',
        'sublist': [1, 2, 3],
        'subdict': {
            'c': 'd'
        }
    },
    'list': [1, True, 'ggg', {
        'h': 'j'
    }, [9, 8, 7]],
}

# The following base classes contain just model declaration for DbAttributes
# and DbExtras and are needed for the methods found at the
# DbAttributeFunctionality and DbExtraFunctionality and used for the deserialization
# of attribute and extras dictionaries
db_attribute_base_model = None
db_extra_base_model = None


class TestAttributesExtrasToJSONMigrationSimple(TestMigrations):
    """
    A "simple" test for the attributes and extra migration from EAV to JSONB.
    It stores a sample dictionary using the EAV deserialization of AiiDA Django
    for the attributes and extras. Then the test checks that they are corerctly
    converted to JSONB.
    """
    migrate_from = '0036_drop_computer_transport_params'
    migrate_to = '0037_attributes_extras_settings_json'

    # In the following dictionary we store the generated nodes (ids, attributes and extras)
    # The correct migration of these nodes will be checked at the test
    nodes_to_verify = dict()

    def setUpBeforeMigration(self):
        global db_attribute_base_model, db_extra_base_model  # pylint: disable=global-statement

        db_node_model = self.apps.get_model('db', 'DbNode')
        db_computer_model = self.apps.get_model('db', 'DbComputer')
        # The following base models are initialized here since the model at this point
        # it has the corresponding EAV tables
        db_attribute_base_model = self.apps.get_model('db', 'DbAttribute')
        db_extra_base_model = self.apps.get_model('db', 'DbExtra')

        computer = db_computer_model(
            name='localhost_migration',
            hostname='localhost',
            transport_type='local',
            scheduler_type='pbspro',
            metadata={'workdir': '/tmp/aiida'}
        )
        computer.save()

        node = db_node_model(node_type='data.Data.', dbcomputer_id=computer.id, user_id=self.default_user.id)
        node.save()

        for key, value in SAMPLE_DICT.items():
            DbAttributeFunctionality.set_value_for_node(node, key, value)

        for key, value in SAMPLE_DICT.items():
            DbExtraFunctionality.set_value_for_node(node, key, value)

        self.nodes_to_verify[node.id] = dict()
        self.nodes_to_verify[node.id]['attr'] = copy.deepcopy(SAMPLE_DICT)
        self.nodes_to_verify[node.id]['extr'] = copy.deepcopy(SAMPLE_DICT)

    def test_attributes_extras_migration(self):
        """Verify that the attributes and extras were migrated correctly"""
        db_node_model = self.apps.get_model('db', 'DbNode')
        for curr_dbnode in db_node_model.objects.all():
            self.assertEqual(curr_dbnode.attributes, self.nodes_to_verify[curr_dbnode.id]['attr'])
            self.assertEqual(curr_dbnode.extras, self.nodes_to_verify[curr_dbnode.id]['extr'])


class TestAttributesExtrasToJSONMigrationManyNodes(TestMigrations):
    """
    This test comparing to the previous one (TestAttributesExtrasToJSONMigrationSimple), it
    creates several nodes with different atributes and extras and checks their correct
    migration one-by-one.
    """
    migrate_from = '0036_drop_computer_transport_params'
    migrate_to = '0037_attributes_extras_settings_json'

    # In the following dictionary we store the generated nodes (ids, attributes and extras)
    # The correct migration of these nodes will be checked at the test
    nodes_to_verify = dict()

    # Number of nodes to create
    nodes_no_to_create = 20

    def setUpBeforeMigration(self):
        global db_attribute_base_model, db_extra_base_model  # pylint: disable=global-statement

        db_node_model = self.apps.get_model('db', 'DbNode')
        db_computer_model = self.apps.get_model('db', 'DbComputer')
        # The following base models are initialized here since the model at this point
        # it has the corresponding EAV tables
        db_attribute_base_model = self.apps.get_model('db', 'DbAttribute')
        db_extra_base_model = self.apps.get_model('db', 'DbExtra')

        computer = db_computer_model(
            name='localhost_migration',
            hostname='localhost',
            transport_type='local',
            scheduler_type='pbspro',
            metadata={'workdir': '/tmp/aiida'}
        )
        computer.save()

        with transaction.atomic():
            for _ in range(self.nodes_no_to_create):
                node = db_node_model(node_type='data.Data.', dbcomputer_id=computer.id, user_id=self.default_user.id)
                node.save()

                attr_copy = copy.deepcopy(SAMPLE_DICT)
                attr_copy['type_of_json'] = 'attr'
                attr_copy['node_id'] = node.id

                # Setting the attributes as it used to be set (with the same methods)
                for key in attr_copy.keys():
                    DbAttributeFunctionality.set_value_for_node(node, key, attr_copy[key])

                extr_copy = copy.deepcopy(SAMPLE_DICT)
                extr_copy['type_of_json'] = 'extr'
                extr_copy['node_id'] = node.id

                # Setting the extras as it used to be set (with the same methods)
                for key in extr_copy.keys():
                    DbExtraFunctionality.set_value_for_node(node, key, extr_copy[key])

                self.nodes_to_verify[node.id] = dict()
                self.nodes_to_verify[node.id]['attr'] = attr_copy
                self.nodes_to_verify[node.id]['extr'] = extr_copy

    def test_attributes_extras_migration_many(self):
        """Verify that the attributes and extras were migrated correctly"""
        db_node_model = self.apps.get_model('db', 'DbNode')
        for curr_dbnode in db_node_model.objects.all():
            self.assertEqual(curr_dbnode.attributes, self.nodes_to_verify[curr_dbnode.id]['attr'])
            self.assertEqual(curr_dbnode.extras, self.nodes_to_verify[curr_dbnode.id]['extr'])


class TestSettingsToJSONMigration(TestMigrations):
    """
    This test checks the correct migration of the settings. Setting records were used as an
    example from a typical settings table of Django EAV.
    """
    migrate_from = '0036_drop_computer_transport_params'
    migrate_to = '0037_attributes_extras_settings_json'

    # The settings to create and verify
    settings_info = dict()

    def setUpBeforeMigration(self):
        from aiida.common import timezone

        db_setting_model = self.apps.get_model('db', 'DbSetting')

        self.settings_info['2daemon|task_stop|updater2'] = dict(
            key='2daemon|task_stop|updater2',
            datatype='date',
            dval=timezone.datetime_to_isoformat(timezone.now()),
            description='The last time the daemon finished to run '
            'the task \'updater\' (updater)'
        )
        self.settings_info['2daemon|task_start|updater2'] = dict(
            key='2daemon|task_start|updater2',
            datatype='date',
            dval=timezone.datetime_to_isoformat(timezone.now()),
            description='The last time the daemon started to run '
            'the task \'updater\' (updater)'
        )
        self.settings_info['2db|backend2'] = dict(
            key='2db|backend2',
            datatype='txt',
            tval='django',
            description='The backend used to communicate with the database.'
        )
        self.settings_info['2daemon|user2'] = dict(
            key='2daemon|user2',
            datatype='txt',
            tval='aiida@theossrv5.epfl.ch',
            description='The only user that is allowed to run the AiiDA daemon on '
            'this DB instance'
        )
        self.settings_info['2db|schemaversion2'] = dict(
            key='2db|schemaversion2',
            datatype='txt',
            tval=' 1.0.8',
            description='The version of the schema used in this database.'
        )

        with transaction.atomic():
            for setting_info in self.settings_info.values():
                setting = db_setting_model(**setting_info)
                setting.save()

    def test_settings_migration(self):
        """Verify that the settings were migrated correctly"""
        db_setting_model = self.apps.get_model('db', 'DbSetting')
        for curr_setting in db_setting_model.objects.filter(key__in=self.settings_info.keys()).all():
            curr_setting_info = self.settings_info[curr_setting.key]
            self.assertEqual(curr_setting.description, curr_setting_info['description'])
            if curr_setting_info['datatype'] == 'txt':
                self.assertEqual(curr_setting.val, curr_setting_info['tval'])
            elif curr_setting_info['datatype'] == 'date':
                self.assertEqual(curr_setting.val, curr_setting_info['dval'])

    def tearDown(self):
        """
        Deletion of settings - this is needed because settings are not deleted by the
        typical test cleanup methods.
        """
        db_setting_model = self.apps.get_model('db', 'DbSetting')
        db_setting_model.objects.filter(key__in=self.settings_info.keys()).delete()
        super(TestSettingsToJSONMigration, self).tearDown()


# pylint: disable=no-init, old-style-class, too-few-public-methods, dangerous-default-value, too-many-statements
# pylint: disable= no-else-return, too-many-arguments, too-many-branches, fixme
class DbMultipleValueAttributeBaseClass():
    """
    Abstract base class for tables storing attribute + value data, of
    different data types (without any association to a Node).
    """
    # separator for subfields
    _sep = '.'  # The AIIDA_ATTRIBUTE_SEP

    class Meta:
        abstract = True
        unique_together = (('key',),)

    # There are no subspecifiers. If instead you want to group attributes
    # (e.g. by node, as it is done in the DbAttributeBaseClass), specify here
    # the field name
    _subspecifier_field_name = None

    @property
    def subspecifier_pk(self):
        """
        Return the subspecifier PK in the database (or None, if no
        subspecifier should be used)
        """
        if self._subspecifier_field_name is None:
            return None
        else:
            return getattr(self, self._subspecifier_field_name).pk

    @classmethod
    def validate_key(cls, key):
        """
        Validate the key string to check if it is valid (e.g., if it does not
        contain the separator symbol.).

        :return: None if the key is valid
        :raise aiida.common.ValidationError: if the key is not valid
        """
        from aiida.backends.utils import validate_attribute_key
        return validate_attribute_key(key)

    @classmethod
    def set_value(
        cls, key, value, with_transaction=True, subspecifier_value=None, other_attribs={}, stop_if_existing=False
    ):
        """
        Set a new value in the DB, possibly associated to the given subspecifier.

        :note: This method also stored directly in the DB.

        :param key: a string with the key to create (must be a level-0
          attribute, that is it cannot contain the separator cls._sep).
        :param value: the value to store (a basic data type or a list or a dict)
        :param subspecifier_value: must be None if this class has no
          subspecifier set (e.g., the DbSetting class).
          Must be the value of the subspecifier (e.g., the dbnode) for classes
          that define it (e.g. DbAttribute and DbExtra)
        :param with_transaction: True if you want this function to be managed
          with transactions. Set to False if you already have a manual
          management of transactions in the block where you are calling this
          function (useful for speed improvements to avoid recursive
          transactions)
        :param other_attribs: a dictionary of other parameters, to store
          only on the level-zero attribute (e.g. for description in DbSetting).
        :param stop_if_existing: if True, it will stop with an
           UniquenessError exception if the new entry would violate an
           uniqueness constraint in the DB (same key, or same key+node,
           depending on the specific subclass). Otherwise, it will
           first delete the old value, if existent. The use with True is
           useful if you want to use a given attribute as a "locking" value,
           e.g. to avoid to perform an action twice on the same node.
           Note that, if you are using transactions, you may get the error
           only when the transaction is committed.
        """
        cls.validate_key(key)

        try:
            if with_transaction:
                sid = transaction.savepoint()

            # create_value returns a list of nodes to store
            to_store = cls.create_value(key, value, subspecifier_value=subspecifier_value, other_attribs=other_attribs)

            if to_store:
                # if not stop_if_existing:
                #     # Delete the olf values if stop_if_existing is False,
                #     # otherwise don't delete them and hope they don't
                #     # exist. If they exist, I'll get an UniquenessError
                #
                #     ## NOTE! Be careful in case the extra/attribute to
                #     ## store is not a simple attribute but a list or dict:
                #     ## like this, it should be ok because if we are
                #     ## overwriting an entry it will stop anyway to avoid
                #     ## to overwrite the main entry, but otherwise
                #     ## there is the risk that trailing pieces remain
                #     ## so in general it is good to recursively clean
                #     ## all sub-items.
                #     cls.del_value(key,
                #                   subspecifier_value=subspecifier_value)
                for my_obj in to_store:
                    my_obj.save()

                # cls.objects.bulk_create(to_store)

            if with_transaction:
                transaction.savepoint_commit(sid)
        except BaseException as exc:  # All exceptions including CTRL+C, ...
            from django.db.utils import IntegrityError
            from aiida.common.exceptions import UniquenessError

            if with_transaction:
                transaction.savepoint_rollback(sid)
            if isinstance(exc, IntegrityError) and stop_if_existing:
                raise UniquenessError(
                    'Impossible to create the required '
                    'entry '
                    "in table '{}', "
                    'another entry already exists and the creation would '
                    'violate an uniqueness constraint.\nFurther details: '
                    '{}'.format(cls.__name__, exc)
                )
            raise

    @classmethod
    def create_value(cls, key, value, subspecifier_value=None, other_attribs={}):
        """
        Create a new list of attributes, without storing them, associated
        with the current key/value pair (and to the given subspecifier,
        e.g. the DbNode for DbAttributes and DbExtras).

        :note: No hits are done on the DB, in particular no check is done
          on the existence of the given nodes.

        :param key: a string with the key to create (can contain the
          separator cls._sep if this is a sub-attribute: indeed, this
          function calls itself recursively)
        :param value: the value to store (a basic data type or a list or a dict)
        :param subspecifier_value: must be None if this class has no
          subspecifier set (e.g., the DbSetting class).
          Must be the value of the subspecifier (e.g., the dbnode) for classes
          that define it (e.g. DbAttribute and DbExtra)
        :param other_attribs: a dictionary of other parameters, to store
          only on the level-zero attribute (e.g. for description in DbSetting).

        :return: always a list of class instances; it is the user
          responsibility to store such entries (typically with a Django
          bulk_create() call).
        """
        import datetime

        from aiida.common import json
        from aiida.common.timezone import is_naive, make_aware, get_current_timezone

        if cls._subspecifier_field_name is None:
            if subspecifier_value is not None:
                raise ValueError(
                    'You cannot specify a subspecifier value for '
                    'class {} because it has no subspecifiers'
                    ''.format(cls.__name__)
                )
            if issubclass(cls, DbAttributeFunctionality):
                new_entry = db_attribute_base_model(key=key, **other_attribs)
            else:
                new_entry = db_extra_base_model(key=key, **other_attribs)
        else:
            if subspecifier_value is None:
                raise ValueError(
                    'You also have to specify a subspecifier value '
                    'for class {} (the {})'.format(cls.__name__, cls._subspecifier_field_name)
                )
            further_params = other_attribs.copy()
            further_params.update({cls._subspecifier_field_name: subspecifier_value})
            # new_entry = cls(key=key, **further_params)
            if issubclass(cls, DbAttributeFunctionality):
                new_entry = db_attribute_base_model(key=key, **further_params)
            else:
                new_entry = db_extra_base_model(key=key, **further_params)

        list_to_return = [new_entry]

        if value is None:
            new_entry.datatype = 'none'
            new_entry.bval = None
            new_entry.tval = ''
            new_entry.ival = None
            new_entry.fval = None
            new_entry.dval = None

        elif isinstance(value, bool):
            new_entry.datatype = 'bool'
            new_entry.bval = value
            new_entry.tval = ''
            new_entry.ival = None
            new_entry.fval = None
            new_entry.dval = None

        elif isinstance(value, six.integer_types):
            new_entry.datatype = 'int'
            new_entry.ival = value
            new_entry.tval = ''
            new_entry.bval = None
            new_entry.fval = None
            new_entry.dval = None

        elif isinstance(value, float):
            new_entry.datatype = 'float'
            new_entry.fval = value
            new_entry.tval = ''
            new_entry.ival = None
            new_entry.bval = None
            new_entry.dval = None

        elif isinstance(value, six.string_types):
            new_entry.datatype = 'txt'
            new_entry.tval = value
            new_entry.bval = None
            new_entry.ival = None
            new_entry.fval = None
            new_entry.dval = None

        elif isinstance(value, datetime.datetime):

            # current timezone is taken from the settings file of django
            if is_naive(value):
                value_to_set = make_aware(value, get_current_timezone())
            else:
                value_to_set = value

            new_entry.datatype = 'date'
            # TODO: time-aware and time-naive datetime objects, see
            # https://docs.djangoproject.com/en/dev/topics/i18n/timezones/#naive-and-aware-datetime-objects
            new_entry.dval = value_to_set
            new_entry.tval = ''
            new_entry.bval = None
            new_entry.ival = None
            new_entry.fval = None

        elif isinstance(value, (list, tuple)):

            new_entry.datatype = 'list'
            new_entry.dval = None
            new_entry.tval = ''
            new_entry.bval = None
            new_entry.ival = len(value)
            new_entry.fval = None

            for i, subv in enumerate(value):
                # I do not need get_or_create here, because
                # above I deleted all children (and I
                # expect no concurrency)
                # NOTE: I do not pass other_attribs
                list_to_return.extend(
                    cls.create_value(
                        key=('{}{}{:d}'.format(key, cls._sep, i)), value=subv, subspecifier_value=subspecifier_value
                    )
                )

        elif isinstance(value, dict):

            new_entry.datatype = 'dict'
            new_entry.dval = None
            new_entry.tval = ''
            new_entry.bval = None
            new_entry.ival = len(value)
            new_entry.fval = None

            for subk, subv in value.items():
                cls.validate_key(subk)

                # I do not need get_or_create here, because
                # above I deleted all children (and I
                # expect no concurrency)
                # NOTE: I do not pass other_attribs
                list_to_return.extend(
                    cls.create_value(
                        key='{}{}{}'.format(key, cls._sep, subk), value=subv, subspecifier_value=subspecifier_value
                    )
                )
        else:
            try:
                jsondata = json.dumps(value)
            except TypeError:
                raise ValueError(
                    'Unable to store the value: it must be either a basic datatype, or json-serializable: {}'.
                    format(value)
                )

            new_entry.datatype = 'json'
            new_entry.tval = jsondata
            new_entry.bval = None
            new_entry.ival = None
            new_entry.fval = None

        return list_to_return


@python_2_unicode_compatible  # pylint: disable=no-init
class DbAttributeBaseClass(DbMultipleValueAttributeBaseClass):
    """
    Abstract base class for tables storing element-attribute-value data.
    Element is the dbnode; attribute is the key name.
    Value is the specific value to store.

    This table had different SQL columns to store different types of data, and
    a datatype field to know the actual datatype.

    Moreover, this class unpacks dictionaries and lists when possible, so that
    it is possible to query inside recursive lists and dicts.
    """

    # In this way, the related name for the DbAttribute inherited class will be
    # 'dbattributes' and for 'dbextra' will be 'dbextras'
    # Moreover, automatically destroy attributes and extras if the parent
    # node is deleted
    # dbnode = m.ForeignKey('DbNode', related_name='%(class)ss', on_delete=m.CASCADE)
    # max_length is required by MySql to have indexes and unique constraints

    _subspecifier_field_name = 'dbnode'

    @classmethod
    def set_value_for_node(cls, dbnode, key, value, with_transaction=True, stop_if_existing=False):
        """
        This is the raw-level method that accesses the DB. No checks are done
        to prevent the user from (re)setting a valid key.
        To be used only internally.

        :todo: there may be some error on concurrent write;
           not checked in this unlucky case!

        :param dbnode: the dbnode for which the attribute should be stored;
          if an integer is passed, it will raise, since this functionality is not
          supported in the models for the migrations.
        :param key: the key of the attribute to store; must be a level-zero
          attribute (i.e., no separators in the key)
        :param value: the value of the attribute to store
        :param with_transaction: if True (default), do this within a transaction,
           so that nothing gets stored if a subitem cannot be created.
           Otherwise, if this parameter is False, no transaction management
           is performed.
        :param stop_if_existing: if True, it will stop with an
           UniquenessError exception if the key already exists
           for the given node. Otherwise, it will
           first delete the old value, if existent. The use with True is
           useful if you want to use a given attribute as a "locking" value,
           e.g. to avoid to perform an action twice on the same node.
           Note that, if you are using transactions, you may get the error
           only when the transaction is committed.

        :raise ValueError: if the key contains the separator symbol used
            internally to unpack dictionaries and lists (defined in cls._sep).
        """
        if isinstance(dbnode, six.integer_types):
            raise ValueError('Integers (the dbnode pk) are not supported as input.')
        else:
            dbnode_node = dbnode

        cls.set_value(
            key,
            value,
            with_transaction=with_transaction,
            subspecifier_value=dbnode_node,
            stop_if_existing=stop_if_existing
        )

    def __str__(self):
        # pylint: disable=no-member
        return '[{} ({})].{} ({})'.format(
            self.dbnode.get_simple_name(invalid_result='Unknown node'),
            self.dbnode.pk,
            self.key,
            self.datatype,
        )


class DbAttributeFunctionality(DbAttributeBaseClass):  # pylint: disable=no-init
    """
    This class defines all the methods that are needed for the correct
    deserialization of given attribute dictionaries to the EAV table.
    It is a stripped-down Django EAV schema to the absolutely necessary
    methods for this deserialization.
    """
    pass  # pylint: disable=unnecessary-pass


class DbExtraFunctionality(DbAttributeBaseClass):  # pylint: disable=no-init
    """
    This class defines all the methods that are needed for the correct
    deserialization of given extras dictionaries to the EAV table.
    It is a stripped-down Django EAV schema to the absolutely necessary
    methods for this deserialization.
    """
    pass  # pylint: disable=unnecessary-pass



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of caching mechanism and configuration for calculations."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import io
import os
import copy
from enum import Enum
from contextlib import contextmanager

import yaml
import six
from wrapt import decorator

from aiida.common import exceptions
from aiida.common.utils import get_object_from_string

__all__ = ('get_use_cache', 'enable_caching', 'disable_caching')


class ConfigKeys(Enum):
    """Valid keys for caching configuration."""

    DEFAULT = 'default'
    ENABLED = 'enabled'
    DISABLED = 'disabled'


DEFAULT_CONFIG = {
    ConfigKeys.DEFAULT.value: False,
    ConfigKeys.ENABLED.value: [],
    ConfigKeys.DISABLED.value: [],
}


def _get_config(config_file):
    """Return the caching configuration.

    :param config_file: the absolute path to the caching configuration file
    :return: the configuration dictionary
    """
    from aiida.manage.configuration import get_profile

    profile = get_profile()

    if profile is None:
        exceptions.ConfigurationError('no profile has been loaded')

    try:
        with io.open(config_file, 'r', encoding='utf8') as handle:
            config = yaml.load(handle)[profile.name]
    # no config file, or no config for this profile
    except (OSError, IOError, KeyError):
        return DEFAULT_CONFIG

    # validate configuration
    for key in config:
        if key not in DEFAULT_CONFIG:
            raise ValueError("Configuration error: Invalid key '{}' in cache_config.yml".format(key))

    # add defaults where config is missing
    for key, default_config in DEFAULT_CONFIG.items():
        config[key] = config.get(key, default_config)

    # load classes
    try:
        for key in [ConfigKeys.ENABLED.value, ConfigKeys.DISABLED.value]:
            config[key] = [get_object_from_string(c) for c in config[key]]
    except (ValueError) as err:
        six.raise_from(
            exceptions.ConfigurationError("Unknown class given in 'cache_config.yml': '{}'".format(err)), err
        )
    return config


_CONFIG = {}


def configure(config_file=None):
    """
    Reads the caching configuration file and sets the _CONFIG variable.
    """
    # pylint: disable=global-statement
    if config_file is None:
        from aiida.manage.configuration import get_config

        config = get_config()
        config_file = os.path.join(config.dirpath, 'cache_config.yml')

    global _CONFIG
    _CONFIG.clear()
    _CONFIG.update(_get_config(config_file=config_file))


@decorator
def _with_config(wrapped, _, args, kwargs):
    """Function decorator to load the caching configuration for the scope of the wrapped function."""
    if not _CONFIG:
        configure()
    return wrapped(*args, **kwargs)


@_with_config
def get_use_cache(node_class=None):
    """Return whether the caching mechanism should be used for the given node class according to the configuration.

    :param node_class: the Node class or sub class to check if enabled for caching
    :return: boolean, True if caching is enabled, False otherwise
    :raises ValueError: if the configuration is invalid by defining the class both enabled and disabled
    """
    if node_class is not None:
        enabled = node_class in _CONFIG[ConfigKeys.ENABLED.value]
        disabled = node_class in _CONFIG[ConfigKeys.DISABLED.value]
        if enabled and disabled:
            raise ValueError('Invalid configuration: Caching for {} is both enabled and disabled.'.format(node_class))
        elif enabled:
            return True
        elif disabled:
            return False
    return _CONFIG[ConfigKeys.DEFAULT.value]


@contextmanager
@_with_config
def _reset_config():
    """Reset the configuration by clearing the contents of the global config variable."""
    # pylint: disable=global-statement
    global _CONFIG
    config_copy = copy.deepcopy(_CONFIG)
    yield
    _CONFIG.clear()
    _CONFIG.update(config_copy)


@contextmanager
def enable_caching(node_class=None):
    """
    Context manager to enable caching, either for a specific node class, or globally.
    Note that this does not affect the behavior of the daemon, only the local Python instance.

    :param node_class: Node class for which caching should be enabled.
    """
    with _reset_config():
        if node_class is None:
            _CONFIG[ConfigKeys.DEFAULT.value] = True
        else:
            _CONFIG[ConfigKeys.ENABLED.value].append(node_class)
            try:
                _CONFIG[ConfigKeys.DISABLED.value].remove(node_class)
            except ValueError:
                pass
        yield


@contextmanager
def disable_caching(node_class=None):
    """
    Context manager to disable caching, either for a specific node class, or globally.
    Note that this does not affect the behavior of the daemon, only the local Python instance.

    :param node_class: Node class for which caching should be disabled.
    """
    with _reset_config():
        if node_class is None:
            _CONFIG[ConfigKeys.DEFAULT.value] = False
        else:
            _CONFIG[ConfigKeys.DISABLED.value].append(node_class)
            try:
                _CONFIG[ConfigKeys.ENABLED.value].remove(node_class)
            except ValueError:
                pass
        yield



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,too-many-statements,too-many-branches
"""`verdi computer` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import io
from functools import partial
from six.moves import zip

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import options, arguments
from aiida.cmdline.params.options.commands import computer as options_computer
from aiida.cmdline.utils import echo
from aiida.cmdline.utils.decorators import with_dbenv
from aiida.cmdline.utils.multi_line_input import ensure_scripts
from aiida.common.exceptions import ValidationError, InputValidationError
from aiida.plugins.entry_point import get_entry_points
from aiida.transports import cli as transport_cli


@verdi.group('computer')
def verdi_computer():
    """Setup and manage computers."""


def get_computer_names():
    """
    Retrieve the list of computers in the DB.
    """
    from aiida.orm.querybuilder import QueryBuilder
    builder = QueryBuilder()
    builder.append(entity_type='computer', project=['name'])
    if builder.count() > 0:
        return next(zip(*builder.all()))  # return the first entry

    return []


def prompt_for_computer_configuration(computer):  # pylint: disable=unused-argument
    pass


def _computer_test_get_jobs(transport, scheduler, authinfo):  # pylint: disable=unused-argument
    """
    Internal test to check if it is possible to check the queue state.

    :param transport: an open transport
    :param scheduler: the corresponding scheduler class
    :param authinfo: the AuthInfo object (from which one can get computer and aiidauser)
    :return: True if the test succeeds, False if it fails.
    """
    echo.echo('> Getting job list...')
    found_jobs = scheduler.get_jobs(as_dict=True)
    echo.echo('  `-> OK, {} jobs found in the queue.'.format(len(found_jobs)))
    return True


def _computer_test_no_unexpected_output(transport, scheduler, authinfo):  # pylint: disable=unused-argument
    """
    Test that there is no unexpected output from the connection.

    This can happen if e.g. there is some spurious command in the
    .bashrc or .bash_profile that is not guarded in case of non-interactive
    shells.

    :param transport: an open transport
    :param scheduler: the corresponding scheduler class
    :param authinfo: the AuthInfo object (from which one can get computer and aiidauser)
    :return: True if the test succeeds, False if it fails.
    """
    # Execute a command that should not return any error
    echo.echo('> Checking that no spurious output is present...')
    retval, stdout, stderr = transport.exec_command_wait('echo -n')
    if retval != 0:
        echo.echo_error("* ERROR! The command 'echo -n' returned a non-zero return code ({})!".format(retval))
        return False
    if stdout:
        echo.echo_error(
            u"""* ERROR! There is some spurious output in the standard output,
that we report below between the === signs:
=========================================================
{}
=========================================================
Please check that you don't have code producing output in
your ~/.bash_profile (or ~/.bashrc). If you don't want to
remove the code, but just to disable it for non-interactive
shells, see comments in issue #1980 on GitHub:
https://github.com/aiidateam/aiida-core/issues/1890
(and in the AiiDA documentation, linked from that issue)
""".format(stdout)
        )
        return False
    if stderr:
        echo.echo_error(
            u"""* ERROR! There is some spurious output in the stderr,
that we report below between the === signs:
=========================================================
{}
=========================================================
Please check that you don't have code producing output in
your ~/.bash_profile (or ~/.bashrc). If you don't want to
remove the code, but just to disable it for non-interactive
shells, see comments in issue #1980 on GitHub:
https://github.com/aiidateam/aiida-core/issues/1890
(and in the AiiDA documentation, linked from that issue)
""".format(stderr)
        )
        return False

    echo.echo('      [OK]')
    return True


def _computer_create_temp_file(transport, scheduler, authinfo):  # pylint: disable=unused-argument
    """
    Internal test to check if it is possible to create a temporary file
    and then delete it in the work directory

    :note: exceptions could be raised

    :param transport: an open transport
    :param scheduler: the corresponding scheduler class
    :param authinfo: the AuthInfo object (from which one can get computer and aiidauser)
    :return: True if the test succeeds, False if it fails.
    """
    import tempfile
    import datetime
    import os

    file_content = "Test from 'verdi computer test' on {}".format(datetime.datetime.now().isoformat())
    echo.echo('> Creating a temporary file in the work directory...')
    echo.echo('  `-> Getting the remote user name...')
    remote_user = transport.whoami()
    echo.echo('      [remote username: {}]'.format(remote_user))
    workdir = authinfo.get_workdir().format(username=remote_user)
    echo.echo('      [Checking/creating work directory: {}]'.format(workdir))

    try:
        transport.chdir(workdir)
    except IOError:
        transport.makedirs(workdir)
        transport.chdir(workdir)

    with tempfile.NamedTemporaryFile(mode='w+') as tempf:
        fname = os.path.split(tempf.name)[1]
        echo.echo('  `-> Creating the file {}...'.format(fname))
        remote_file_path = os.path.join(workdir, fname)
        tempf.write(file_content)
        tempf.flush()
        transport.putfile(tempf.name, remote_file_path)
    echo.echo('  `-> Checking if the file has been created...')
    if not transport.path_exists(remote_file_path):
        echo.echo_error('* ERROR! The file was not found!')
        return False

    echo.echo('      [OK]')
    echo.echo('  `-> Retrieving the file and checking its content...')

    handle, destfile = tempfile.mkstemp()
    os.close(handle)
    try:
        transport.getfile(remote_file_path, destfile)
        with io.open(destfile, encoding='utf8') as dfile:
            read_string = dfile.read()
        echo.echo('      [Retrieved]')
        if read_string != file_content:
            echo.echo_error('* ERROR! The file content is different from what was expected!')
            echo.echo('** Expected:')
            echo.echo(file_content)
            echo.echo('** Found:')
            echo.echo(read_string)
            return False

        echo.echo('      [Content OK]')
    finally:
        os.remove(destfile)

    echo.echo('  `-> Removing the file...')
    transport.remove(remote_file_path)
    echo.echo('  [Deleted successfully]')
    return True


def get_parameter_default(parameter, ctx):
    """
    Get the value for a specific parameter from the computer_builder or the default value of that option

    :param parameter: parameter name
    :param ctx: click context of the command
    :return: parameter default value, or None
    """
    default = None

    for param in ctx.command.get_params(ctx):
        if param.name == parameter:
            default = param.default

    try:
        value = getattr(ctx.computer_builder, parameter)
        if value == '' or value is None:
            value = default
    except KeyError:
        value = default

    return value


# pylint: disable=unused-argument
def set_computer_builder(ctx, param, value):
    """Set the computer spec for defaults of following options."""
    from aiida.orm.utils.builders.computer import ComputerBuilder
    ctx.computer_builder = ComputerBuilder.from_computer(value)
    return value


@verdi_computer.command('setup')
@options_computer.LABEL()
@options_computer.HOSTNAME()
@options_computer.DESCRIPTION()
@options_computer.TRANSPORT()
@options_computer.SCHEDULER()
@options_computer.SHEBANG()
@options_computer.WORKDIR()
@options_computer.MPI_RUN_COMMAND()
@options_computer.MPI_PROCS_PER_MACHINE()
@options.PREPEND_TEXT()
@options.APPEND_TEXT()
@options.NON_INTERACTIVE()
@options.CONFIG_FILE()
@click.pass_context
@with_dbenv()
def computer_setup(ctx, non_interactive, **kwargs):
    """Create a new computer."""
    from aiida.orm.utils.builders.computer import ComputerBuilder

    if kwargs['label'] in get_computer_names():
        echo.echo_critical(
            'A computer called {c} already exists. '
            'Use "verdi computer duplicate {c}" to set up a new '
            'computer starting from the settings of {c}.'.format(c=kwargs['label'])
        )

    if not non_interactive:
        try:
            pre, post = ensure_scripts(kwargs.pop('prepend_text', ''), kwargs.pop('append_text', ''), kwargs)
        except InputValidationError as exception:
            raise click.BadParameter('invalid prepend and or append text: {}'.format(exception))

        kwargs['prepend_text'] = pre
        kwargs['append_text'] = post

    kwargs['transport'] = kwargs['transport'].name
    kwargs['scheduler'] = kwargs['scheduler'].name

    computer_builder = ComputerBuilder(**kwargs)
    try:
        computer = computer_builder.new()
    except (ComputerBuilder.ComputerValidationError, ValidationError) as e:
        echo.echo_critical('{}: {}'.format(type(e).__name__, e))

    try:
        computer.store()
    except ValidationError as err:
        echo.echo_critical('unable to store the computer: {}. Exiting...'.format(err))
    else:
        echo.echo_success('Computer<{}> {} created'.format(computer.pk, computer.name))

    echo.echo_info('Note: before the computer can be used, it has to be configured with the command:')
    echo.echo_info('  verdi computer configure {} {}'.format(computer.get_transport_type(), computer.name))


@verdi_computer.command('duplicate')
@arguments.COMPUTER(callback=set_computer_builder)
@options_computer.LABEL(contextual_default=partial(get_parameter_default, 'label'))
@options_computer.HOSTNAME(contextual_default=partial(get_parameter_default, 'hostname'))
@options_computer.DESCRIPTION(contextual_default=partial(get_parameter_default, 'description'))
@options_computer.TRANSPORT(contextual_default=partial(get_parameter_default, 'transport'))
@options_computer.SCHEDULER(contextual_default=partial(get_parameter_default, 'scheduler'))
@options_computer.SHEBANG(contextual_default=partial(get_parameter_default, 'shebang'))
@options_computer.WORKDIR(contextual_default=partial(get_parameter_default, 'work_dir'))
@options_computer.MPI_RUN_COMMAND(contextual_default=partial(get_parameter_default, 'mpirun_command'))
@options_computer.MPI_PROCS_PER_MACHINE(contextual_default=partial(get_parameter_default, 'mpiprocs_per_machine'))
@options.PREPEND_TEXT()
@options.APPEND_TEXT()
@options.NON_INTERACTIVE()
@click.pass_context
@with_dbenv()
def computer_duplicate(ctx, computer, non_interactive, **kwargs):
    """Duplicate a computer allowing to change some parameters."""
    from aiida import orm
    from aiida.orm.utils.builders.computer import ComputerBuilder

    if kwargs['label'] in get_computer_names():
        echo.echo_critical('A computer called {} already exists'.format(kwargs['label']))

    if not non_interactive:
        try:
            pre, post = ensure_scripts(kwargs.pop('prepend_text', ''), kwargs.pop('append_text', ''), kwargs)
        except InputValidationError as exception:
            raise click.BadParameter('invalid prepend and or append text: {}'.format(exception))

        kwargs['prepend_text'] = pre
        kwargs['append_text'] = post

    kwargs['transport'] = kwargs['transport'].name
    kwargs['scheduler'] = kwargs['scheduler'].name

    computer_builder = ctx.computer_builder
    for key, value in kwargs.items():
        if value is not None:
            setattr(computer_builder, key, value)

    try:
        computer = computer_builder.new()
    except (ComputerBuilder.ComputerValidationError, ValidationError) as e:
        echo.echo_critical('{}: {}'.format(type(e).__name__, e))
    else:
        echo.echo_success('stored computer {}<{}>'.format(computer.name, computer.pk))

    try:
        computer.store()
    except ValidationError as err:
        echo.echo_critical('unable to store the computer: {}. Exiting...'.format(err))
    else:
        echo.echo_success('Computer<{}> {} created'.format(computer.pk, computer.name))

    is_configured = computer.is_user_configured(orm.User.objects.get_default())

    if not is_configured:
        echo.echo_info('Note: before the computer can be used, it has to be configured with the command:')
        echo.echo_info('  verdi computer configure {} {}'.format(computer.get_transport_type(), computer.name))


@verdi_computer.command('enable')
@arguments.COMPUTER()
@arguments.USER()
@with_dbenv()
def computer_enable(computer, user):
    """Enable the computer for the given user."""
    from aiida.common.exceptions import NotExistent

    try:
        authinfo = computer.get_authinfo(user)
    except NotExistent:
        echo.echo_critical(
            "User with email '{}' is not configured for computer '{}' yet.".format(user.email, computer.name)
        )

    if not authinfo.enabled:
        authinfo.enabled = True
        echo.echo_info("Computer '{}' enabled for user {}.".format(computer.name, user.get_full_name()))
    else:
        echo.echo_info(
            "Computer '{}' was already enabled for user {} {}.".format(computer.name, user.first_name, user.last_name)
        )


@verdi_computer.command('disable')
@arguments.COMPUTER()
@arguments.USER()
@with_dbenv()
def computer_disable(computer, user):
    """Disable the computer for the given user.

    Thi can be useful, for example, when a computer is under maintenance."""
    from aiida.common.exceptions import NotExistent

    try:
        authinfo = computer.get_authinfo(user)
    except NotExistent:
        echo.echo_critical(
            "User with email '{}' is not configured for computer '{}' yet.".format(user.email, computer.name)
        )

    if authinfo.enabled:
        authinfo.enabled = False
        echo.echo_info("Computer '{}' disabled for user {}.".format(computer.name, user.get_full_name()))
    else:
        echo.echo_info(
            "Computer '{}' was already disabled for user {} {}.".format(computer.name, user.first_name, user.last_name)
        )


@verdi_computer.command('list')
@options.ALL(help='Show also disabled or unconfigured computers.')
@options.RAW(help='Show only the computer names, one per line.')
@with_dbenv()
def computer_list(all_entries, raw):
    """List all available computers."""
    from aiida.orm import Computer, User

    if not raw:
        echo.echo_info('List of configured computers')
        echo.echo_info("Use 'verdi computer show COMPUTERNAME' to display more detailed information")

    computers = Computer.objects.all()
    user = User.objects.get_default()

    if not computers:
        echo.echo_info("No computers configured yet. Use 'verdi computer setup'")

    sort = lambda computer: computer.name
    highlight = lambda comp: comp.is_user_configured(user) and comp.is_user_enabled(user)
    hide = lambda comp: not (comp.is_user_configured(user) and comp.is_user_enabled(user)) and not all_entries
    echo.echo_formatted_list(computers, ['name'], sort=sort, highlight=highlight, hide=hide)


@verdi_computer.command('show')
@arguments.COMPUTER()
@with_dbenv()
def computer_show(computer):
    """Show detailed information for a computer."""
    echo.echo(computer.full_text_info)


@verdi_computer.command('rename')
@arguments.COMPUTER()
@arguments.LABEL('NEW_NAME')
@with_dbenv()
def computer_rename(computer, new_name):
    """Rename a computer."""
    from aiida.common.exceptions import UniquenessError

    old_name = computer.get_name()

    if old_name == new_name:
        echo.echo_critical('The old and new names are the same.')

    try:
        computer.set_name(new_name)
        computer.store()
    except ValidationError as error:
        echo.echo_critical('Invalid input! {}'.format(error))
    except UniquenessError as error:
        echo.echo_critical(
            'Uniqueness error encountered! Probably a '
            "computer with name '{}' already exists"
            ''.format(new_name)
        )
        echo.echo_critical('(Message was: {})'.format(error))

    echo.echo_success("Computer '{}' renamed to '{}'".format(old_name, new_name))


@verdi_computer.command('test')
@options.USER(
    required=False,
    help='Test the connection for a given AiiDA user, specified by'
    'their email address. If not specified, uses the current default user.',
)
@click.option(
    '-t',
    '--print-traceback',
    is_flag=True,
    help='Print the full traceback in case an exception is raised',
)
@arguments.COMPUTER()
@with_dbenv()
def computer_test(user, print_traceback, computer):
    """
    Test the connection to a computer.

    It tries to connect, to get the list of calculations on the queue and
    to perform other tests.
    """
    import traceback
    from aiida.common.exceptions import NotExistent
    from aiida import orm

    # Set a user automatically if one is not specified in the command line
    if user is None:
        user = orm.User.objects.get_default()

    echo.echo('Testing computer<{}> for user<{}>...'.format(computer.name, user.email))

    try:
        authinfo = computer.get_authinfo(user)
    except NotExistent:
        echo.echo_critical('Computer<{}> is not yet configured for user<{}>'.format(computer.name, user.email))

    if not authinfo.enabled:
        echo.echo_warning('Computer<{}> is disabled for user<{}>'.format(computer.name, user.email))
        click.confirm('Do you really want to test it?', abort=True)

    scheduler = authinfo.computer.get_scheduler()
    transport = authinfo.get_transport()

    # STARTING TESTS HERE
    num_failures = 0
    num_tests = 0

    try:
        echo.echo('> Testing connection...')
        with transport:
            scheduler.set_transport(transport)
            num_tests += 1
            for test in [_computer_test_no_unexpected_output, _computer_test_get_jobs, _computer_create_temp_file]:
                num_tests += 1
                try:
                    succeeded = test(transport=transport, scheduler=scheduler, authinfo=authinfo)
                # pylint:disable=broad-except
                except Exception as error:
                    echo.echo_error('* The test raised an exception!')
                    if print_traceback:
                        echo.echo('** Full traceback:')
                        # Indent
                        echo.echo('\n'.join(['   {}'.format(l) for l in traceback.format_exc().splitlines()]))
                    else:
                        echo.echo('** {}: {}'.format(error.__class__.__name__, error))
                        echo.echo('** (use the --print-traceback option to see the full traceback)')
                    succeeded = False

                if not succeeded:
                    num_failures += 1

        if num_failures:
            echo.echo('Some tests failed! ({} out of {} failed)'.format(num_failures, num_tests))
        else:
            echo.echo('Test completed (all {} tests succeeded)'.format(num_tests))
    except Exception as error:  # pylint:disable=broad-except
        echo.echo_error('** Error while trying to connect to the computer! Cannot perform following tests, stopping.')
        if print_traceback:
            echo.echo('** Full traceback:')
            echo.echo('\n'.join(['   {}'.format(l) for l in traceback.format_exc().splitlines()]))
        else:
            echo.echo('{}: {}'.format(error.__class__.__name__, error))
            echo.echo('(use the --print-traceback option to see the full traceback)')
        succeeded = False


@verdi_computer.command('delete')
@arguments.COMPUTER()
@with_dbenv()
def computer_delete(computer):
    """
    Delete a computer.

    Note that it is not possible to delete the computer if there are calculations that are using it.
    """
    from aiida.common.exceptions import InvalidOperation
    from aiida import orm

    compname = computer.name

    try:
        orm.Computer.objects.delete(computer.id)
    except InvalidOperation as error:
        echo.echo_critical(str(error))

    echo.echo_success("Computer '{}' deleted.".format(compname))


@verdi_computer.group('configure')
def computer_configure():
    """Configure the Authinfo details for a computer (and user)."""


@computer_configure.command('show')
@click.option(
    '--defaults', is_flag=True, default=False, help='Show the default configuration settings for this computer.'
)
@click.option('--as-option-string', is_flag=True)
@options.USER(
    help='Email address of the AiiDA user for whom to configure this computer (if different from default user).'
)
@arguments.COMPUTER()
def computer_config_show(computer, user, defaults, as_option_string):
    """Show the current configuration for a computer."""
    import tabulate
    from aiida.common.escaping import escape_for_bash

    transport_cls = computer.get_transport_class()
    option_list = [
        param for param in transport_cli.create_configure_cmd(computer.get_transport_type()).params
        if isinstance(param, click.core.Option)
    ]
    option_list = [option for option in option_list if option.name in transport_cls.get_valid_auth_params()]

    if defaults:
        config = {option.name: transport_cli.transport_option_default(option.name, computer) for option in option_list}
    else:
        config = computer.get_configuration(user)

    option_items = []
    if as_option_string:
        for option in option_list:
            t_opt = transport_cls.auth_options[option.name]
            if config.get(option.name) or config.get(option.name) is False:
                if t_opt.get('switch'):
                    option_value = option.opts[-1] if config.get(option.name
                                                                ) else '--no-{}'.format(option.name.replace('_', '-'))
                elif t_opt.get('is_flag'):
                    is_default = config.get(option.name
                                           ) == transport_cli.transport_option_default(option.name, computer)
                    option_value = option.opts[-1] if is_default else ''
                else:
                    option_value = '{}={}'.format(option.opts[-1], option.type(config[option.name]))
                option_items.append(option_value)
        opt_string = ' '.join(option_items)
        echo.echo(escape_for_bash(opt_string))
    else:
        table = []
        for name in transport_cls.get_valid_auth_params():
            if name in config:
                table.append(('* ' + name, config[name]))
            else:
                table.append(('* ' + name, '-'))
        echo.echo(tabulate.tabulate(table, tablefmt='plain'))


for ep in get_entry_points('aiida.transports'):
    computer_configure.add_command(transport_cli.create_configure_cmd(ep.name))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the process_function decorator."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.engine import run, run_get_node, submit, calcfunction, workfunction, Process, ExitCode
from aiida.orm.nodes.data.bool import get_true_node

DEFAULT_INT = 256
DEFAULT_LABEL = 'Default label'
DEFAULT_DESCRIPTION = 'Default description'
CUSTOM_LABEL = 'Custom label'
CUSTOM_DESCRIPTION = 'Custom description'


class TestProcessFunction(AiidaTestCase):
    """
    Note that here we use `@workfunctions` and `@calculations`, the concrete versions of the
    `@process_function` decorator, even though we are testing only the shared functionality
    that is captured in the `@process_function` decorator, relating to the transformation
    of the wrapped function into a `FunctionProcess`.
    The reason we do not use the `@process_function` decorator itself, is because it
    does not have a node class by default. We could create one on the fly, but then
    anytime inputs or outputs would be attached to it in the tests, the `validate_link`
    function would complain as the dummy node class is not recognized as a valid process node.
    """

    # pylint: disable=too-many-public-methods,too-many-instance-attributes

    def setUp(self):
        super(TestProcessFunction, self).setUp()
        self.assertIsNone(Process.current())

        @workfunction
        def function_return_input(data):
            return data

        @calcfunction
        def function_return_true():
            return get_true_node()

        @workfunction
        def function_args(data_a):
            return data_a

        @workfunction
        def function_args_with_default(data_a=orm.Int(DEFAULT_INT)):
            return data_a

        @calcfunction
        def function_with_none_default(int_a, int_b, int_c=None):
            if int_c is not None:
                return orm.Int(int_a + int_b + int_c)
            return orm.Int(int_a + int_b)

        @workfunction
        def function_kwargs(**kwargs):
            return kwargs

        @workfunction
        def function_args_and_kwargs(data_a, **kwargs):
            result = {'data_a': data_a}
            result.update(kwargs)
            return result

        @workfunction
        def function_args_and_default(data_a, data_b=orm.Int(DEFAULT_INT)):
            return {'data_a': data_a, 'data_b': data_b}

        @workfunction
        def function_defaults(
            data_a=orm.Int(DEFAULT_INT), metadata={
                'label': DEFAULT_LABEL,
                'description': DEFAULT_DESCRIPTION
            }
        ):  # pylint: disable=unused-argument,dangerous-default-value,missing-docstring
            return data_a

        @workfunction
        def function_default_label():
            return

        @workfunction
        def function_exit_code(exit_status, exit_message):
            return ExitCode(exit_status.value, exit_message.value)

        @workfunction
        def function_excepts(exception):
            raise RuntimeError(exception.value)

        @workfunction
        def function_out_unstored():
            return orm.Int(DEFAULT_INT)

        self.function_return_input = function_return_input
        self.function_return_true = function_return_true
        self.function_args = function_args
        self.function_args_with_default = function_args_with_default
        self.function_with_none_default = function_with_none_default
        self.function_kwargs = function_kwargs
        self.function_args_and_kwargs = function_args_and_kwargs
        self.function_args_and_default = function_args_and_default
        self.function_defaults = function_defaults
        self.function_default_label = function_default_label
        self.function_exit_code = function_exit_code
        self.function_excepts = function_excepts
        self.function_out_unstored = function_out_unstored

    def tearDown(self):
        super(TestProcessFunction, self).tearDown()
        self.assertIsNone(Process.current())

    def test_plugin_version(self):
        """Test the version attributes of a process function."""
        from aiida import __version__ as version_core

        _, node = self.function_args_with_default.run_get_node()

        # Since the "plugin" i.e. the process function is defined in `aiida-core` the `version.plugin` is the same as
        # the version of `aiida-core` itself
        version_info = node.get_attribute('version')
        self.assertEqual(version_info['core'], version_core)
        self.assertEqual(version_info['plugin'], version_core)

    def test_process_state(self):
        """Test the process state for a process function."""
        _, node = self.function_args_with_default.run_get_node()

        self.assertEqual(node.is_terminated, True)
        self.assertEqual(node.is_excepted, False)
        self.assertEqual(node.is_killed, False)
        self.assertEqual(node.is_finished, True)
        self.assertEqual(node.is_finished_ok, True)
        self.assertEqual(node.is_failed, False)

    def test_process_type(self):
        """Test that the process type correctly contains the module and name of original decorated function."""
        _, node = self.function_defaults.run_get_node()
        process_type = '{}.{}'.format(self.function_defaults.__module__, self.function_defaults.__name__)
        self.assertEqual(node.process_type, process_type)

    def test_exit_status(self):
        """A FINISHED process function has to have an exit status of 0"""
        _, node = self.function_args_with_default.run_get_node()
        self.assertEqual(node.exit_status, 0)
        self.assertEqual(node.is_finished_ok, True)
        self.assertEqual(node.is_failed, False)

    def test_source_code_attributes(self):
        """Verify function properties are properly introspected and stored in the nodes attributes and repository."""
        function_name = 'test_process_function'

        @calcfunction
        def test_process_function(data):
            return {'result': orm.Int(data.value + 1)}

        _, node = test_process_function.run_get_node(data=orm.Int(5))

        # Read the source file of the calculation function that should be stored in the repository
        function_source_code = node.get_function_source_code().split('\n')

        # Verify that the function name is correct and the first source code linenumber is stored
        self.assertEqual(node.function_name, function_name)
        self.assertIsInstance(node.function_starting_line_number, int)

        # Check that first line number is correct. Note that the first line should correspond
        # to the `@workfunction` directive, but since the list is zero-indexed we actually get the
        # following line, which should correspond to the function name i.e. `def test_process_function(data)`
        function_name_from_source = function_source_code[node.function_starting_line_number]
        self.assertTrue(node.function_name in function_name_from_source)

    def test_function_varargs(self):
        """Variadic arguments are not supported and should raise."""
        with self.assertRaises(ValueError):

            @workfunction
            def function_varargs(*args):  # pylint: disable=unused-variable
                return args

    def test_function_args(self):
        """Simple process function that defines a single positional argument."""
        arg = 1

        with self.assertRaises(ValueError):
            result = self.function_args()  # pylint: disable=no-value-for-parameter

        result = self.function_args(data_a=orm.Int(arg))
        self.assertTrue(isinstance(result, orm.Int))
        self.assertEqual(result, arg)

    def test_function_args_with_default(self):
        """Simple process function that defines a single argument with a default."""
        arg = 1

        result = self.function_args_with_default()
        self.assertTrue(isinstance(result, orm.Int))
        self.assertEqual(result, orm.Int(DEFAULT_INT))

        result = self.function_args_with_default(data_a=orm.Int(arg))
        self.assertTrue(isinstance(result, orm.Int))
        self.assertEqual(result, arg)

    def test_function_with_none_default(self):
        """Simple process function that defines a keyword with `None` as default value."""
        int_a = orm.Int(1)
        int_b = orm.Int(2)
        int_c = orm.Int(3)

        result = self.function_with_none_default(int_a, int_b)
        self.assertTrue(isinstance(result, orm.Int))
        self.assertEqual(result, orm.Int(3))

        result = self.function_with_none_default(int_a, int_b, int_c)
        self.assertTrue(isinstance(result, orm.Int))
        self.assertEqual(result, orm.Int(6))

    def test_function_kwargs(self):
        """Simple process function that defines keyword arguments."""
        kwargs = {'data_a': orm.Int(DEFAULT_INT)}

        result, node = self.function_kwargs.run_get_node()
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(len(node.get_incoming().all()), 0)
        self.assertEqual(result, {})

        result, node = self.function_kwargs.run_get_node(**kwargs)
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(len(node.get_incoming().all()), 1)
        self.assertEqual(result, kwargs)

        # Calling with any number of positional arguments should raise
        with self.assertRaises(TypeError):
            self.function_kwargs.run_get_node(orm.Int(1))

        with self.assertRaises(TypeError):
            self.function_kwargs.run_get_node(orm.Int(1), b=orm.Int(2))

    def test_function_args_and_kwargs(self):
        """Simple process function that defines a positional argument and keyword arguments."""
        arg = 1
        args = (orm.Int(DEFAULT_INT),)
        kwargs = {'data_b': orm.Int(arg)}

        result = self.function_args_and_kwargs(*args)
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result, {'data_a': args[0]})

        result = self.function_args_and_kwargs(*args, **kwargs)
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result, {'data_a': args[0], 'data_b': kwargs['data_b']})

        # Calling with more positional arguments than defined in the signature should raise
        with self.assertRaises(TypeError):
            self.function_kwargs.run_get_node(orm.Int(1), orm.Int(2))

        with self.assertRaises(TypeError):
            self.function_kwargs.run_get_node(orm.Int(1), orm.Int(2), b=orm.Int(2))

    def test_function_args_and_kwargs_default(self):
        """Simple process function that defines a positional argument and an argument with a default."""
        arg = 1
        args_input_default = (orm.Int(DEFAULT_INT),)
        args_input_explicit = (orm.Int(DEFAULT_INT), orm.Int(arg))

        result = self.function_args_and_default(*args_input_default)
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result, {'data_a': args_input_default[0], 'data_b': orm.Int(DEFAULT_INT)})

        result = self.function_args_and_default(*args_input_explicit)
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result, {'data_a': args_input_explicit[0], 'data_b': args_input_explicit[1]})

    def test_function_args_passing_kwargs(self):
        """Cannot pass kwargs if the function does not explicitly define it accepts kwargs."""
        arg = 1

        with self.assertRaises(ValueError):
            self.function_args(data_a=orm.Int(arg), data_b=orm.Int(arg))  # pylint: disable=unexpected-keyword-arg

    def test_function_set_label_description(self):
        """Verify that the label and description can be set for all process function variants."""
        metadata = {'label': CUSTOM_LABEL, 'description': CUSTOM_DESCRIPTION}

        _, node = self.function_args.run_get_node(data_a=orm.Int(DEFAULT_INT), metadata=metadata)
        self.assertEqual(node.label, CUSTOM_LABEL)
        self.assertEqual(node.description, CUSTOM_DESCRIPTION)

        _, node = self.function_args_with_default.run_get_node(metadata=metadata)
        self.assertEqual(node.label, CUSTOM_LABEL)
        self.assertEqual(node.description, CUSTOM_DESCRIPTION)

        _, node = self.function_kwargs.run_get_node(metadata=metadata)
        self.assertEqual(node.label, CUSTOM_LABEL)
        self.assertEqual(node.description, CUSTOM_DESCRIPTION)

        _, node = self.function_args_and_kwargs.run_get_node(data_a=orm.Int(DEFAULT_INT), metadata=metadata)
        self.assertEqual(node.label, CUSTOM_LABEL)
        self.assertEqual(node.description, CUSTOM_DESCRIPTION)

        _, node = self.function_args_and_default.run_get_node(data_a=orm.Int(DEFAULT_INT), metadata=metadata)
        self.assertEqual(node.label, CUSTOM_LABEL)
        self.assertEqual(node.description, CUSTOM_DESCRIPTION)

    def test_function_defaults(self):
        """Verify that a process function can define a default label and description but can be overriden."""
        metadata = {'label': CUSTOM_LABEL, 'description': CUSTOM_DESCRIPTION}

        _, node = self.function_defaults.run_get_node(data_a=orm.Int(DEFAULT_INT))
        self.assertEqual(node.label, DEFAULT_LABEL)
        self.assertEqual(node.description, DEFAULT_DESCRIPTION)

        _, node = self.function_defaults.run_get_node(metadata=metadata)
        self.assertEqual(node.label, CUSTOM_LABEL)
        self.assertEqual(node.description, CUSTOM_DESCRIPTION)

    def test_function_default_label(self):
        """Verify unless specified label is taken from function name."""
        metadata = {'label': CUSTOM_LABEL, 'description': CUSTOM_DESCRIPTION}

        _, node = self.function_default_label.run_get_node()
        self.assertEqual(node.label, 'function_default_label')
        self.assertEqual(node.description, '')

        _, node = self.function_default_label.run_get_node(metadata=metadata)
        self.assertEqual(node.label, CUSTOM_LABEL)
        self.assertEqual(node.description, CUSTOM_DESCRIPTION)

    def test_launchers(self):
        """Verify that the various launchers are working."""
        result = run(self.function_return_true)
        self.assertTrue(result)

        result, node = run_get_node(self.function_return_true)
        self.assertTrue(result)
        self.assertEqual(result, get_true_node())
        self.assertTrue(isinstance(node, orm.CalcFunctionNode))

        with self.assertRaises(AssertionError):
            submit(self.function_return_true)

    def test_return_exit_code(self):
        """
        A process function that returns an ExitCode namedtuple should have its exit status and message set FINISHED
        """
        exit_status = 418
        exit_message = 'I am a teapot'

        message = orm.Str(exit_message)
        _, node = self.function_exit_code.run_get_node(exit_status=orm.Int(exit_status), exit_message=message)

        self.assertTrue(node.is_finished)
        self.assertFalse(node.is_finished_ok)
        self.assertEqual(node.exit_status, exit_status)
        self.assertEqual(node.exit_message, exit_message)

    def test_normal_exception(self):
        """If a process, for example a FunctionProcess, excepts, the exception should be stored in the node."""
        exception = 'This process function excepted'

        with self.assertRaises(RuntimeError):
            _, node = self.function_excepts.run_get_node(exception=orm.Str(exception))
            self.assertTrue(node.is_excepted)
            self.assertEqual(node.exception, exception)

    def test_function_out_unstored(self):
        """A workfunction that returns an unstored node should raise as it indicates users tried to create data."""
        with self.assertRaises(ValueError):
            self.function_out_unstored()

    def test_simple_workflow(self):
        """Test construction of simple workflow by chaining process functions."""

        @calcfunction
        def add(data_a, data_b):
            return data_a + data_b

        @calcfunction
        def mul(data_a, data_b):
            return data_a * data_b

        @workfunction
        def add_mul_wf(data_a, data_b, data_c):
            return mul(add(data_a, data_b), data_c)

        result, node = add_mul_wf.run_get_node(orm.Int(3), orm.Int(4), orm.Int(5))

        self.assertEqual(result, (3 + 4) * 5)
        self.assertIsInstance(node, orm.WorkFunctionNode)

    def test_hashes(self):
        """Test that the hashes generated for identical process functions with identical inputs are the same."""
        _, node1 = self.function_return_input.run_get_node(data=orm.Int(2))
        _, node2 = self.function_return_input.run_get_node(data=orm.Int(2))
        self.assertEqual(node1.get_hash(), node1.get_extra('_aiida_hash'))
        self.assertEqual(node2.get_hash(), node2.get_extra('_aiida_hash'))
        self.assertEqual(node1.get_hash(), node2.get_hash())

    def test_hashes_different(self):
        """Test that the hashes generated for identical process functions with different inputs are the different."""
        _, node1 = self.function_return_input.run_get_node(data=orm.Int(2))
        _, node2 = self.function_return_input.run_get_node(data=orm.Int(3))
        self.assertEqual(node1.get_hash(), node1.get_extra('_aiida_hash'))
        self.assertEqual(node2.get_hash(), node2.get_extra('_aiida_hash'))
        self.assertNotEqual(node1.get_hash(), node2.get_hash())



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""AiiDA profile related code"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os
from six import PY2

if PY2:
    import collections
else:
    import collections.abc as collections  # pylint: disable=import-error, no-name-in-module

from aiida.common import exceptions  # pylint: disable=wrong-import-position
from .settings import DAEMON_DIR, DAEMON_LOG_DIR  # pylint: disable=wrong-import-position

__all__ = ('Profile',)

CIRCUS_PID_FILE_TEMPLATE = os.path.join(DAEMON_DIR, 'circus-{}.pid')
DAEMON_PID_FILE_TEMPLATE = os.path.join(DAEMON_DIR, 'aiida-{}.pid')
CIRCUS_LOG_FILE_TEMPLATE = os.path.join(DAEMON_LOG_DIR, 'circus-{}.log')
DAEMON_LOG_FILE_TEMPLATE = os.path.join(DAEMON_LOG_DIR, 'aiida-{}.log')
CIRCUS_PORT_FILE_TEMPLATE = os.path.join(DAEMON_DIR, 'circus-{}.port')
CIRCUS_SOCKET_FILE_TEMPATE = os.path.join(DAEMON_DIR, 'circus-{}.sockets')
CIRCUS_CONTROLLER_SOCKET_TEMPLATE = 'circus.c.sock'
CIRCUS_PUBSUB_SOCKET_TEMPLATE = 'circus.p.sock'
CIRCUS_STATS_SOCKET_TEMPLATE = 'circus.s.sock'


class Profile(object):  # pylint: disable=too-many-public-methods
    """Class that models a profile as it is stored in the configuration file of an AiiDA instance."""

    RMQ_PREFIX = 'aiida-{uuid}'
    KEY_OPTIONS = 'options'
    KEY_UUID = 'PROFILE_UUID'
    KEY_DEFAULT_USER = 'default_user_email'
    KEY_DATABASE_ENGINE = 'AIIDADB_ENGINE'
    KEY_DATABASE_BACKEND = 'AIIDADB_BACKEND'
    KEY_DATABASE_NAME = 'AIIDADB_NAME'
    KEY_DATABASE_PORT = 'AIIDADB_PORT'
    KEY_DATABASE_HOSTNAME = 'AIIDADB_HOST'
    KEY_DATABASE_USERNAME = 'AIIDADB_USER'
    KEY_DATABASE_PASSWORD = 'AIIDADB_PASS'  # noqa
    KEY_REPOSITORY_URI = 'AIIDADB_REPOSITORY_URI'

    # A mapping of valid attributes to the key under which they are stored in the configuration dictionary
    _map_config_to_internal = {
        KEY_OPTIONS: 'options',
        KEY_UUID: 'uuid',
        KEY_DEFAULT_USER: 'default_user',
        KEY_DATABASE_ENGINE: 'database_engine',
        KEY_DATABASE_BACKEND: 'database_backend',
        KEY_DATABASE_NAME: 'database_name',
        KEY_DATABASE_PORT: 'database_port',
        KEY_DATABASE_HOSTNAME: 'database_hostname',
        KEY_DATABASE_USERNAME: 'database_username',
        KEY_DATABASE_PASSWORD: 'database_password',
        KEY_REPOSITORY_URI: 'repository_uri',
    }

    @classmethod
    def contains_unknown_keys(cls, dictionary):
        """Return whether the profile dictionary contains any unsupported keys.

        :param dictionary: a profile dictionary
        :return: boolean, True when the dictionay contains unsupported keys
        """
        return set(dictionary.keys()) - set(cls._map_config_to_internal.keys())

    def __init__(self, name, attributes, from_config=False):
        if not isinstance(attributes, collections.Mapping):
            raise TypeError('attributes should be a mapping but is {}'.format(type(attributes)))

        self._name = name
        self._attributes = {}

        for internal_key, value in attributes.items():
            if from_config:
                try:
                    internal_key = self._map_config_to_internal[internal_key]
                except KeyError:
                    from aiida.cmdline.utils import echo
                    echo.echo_warning(
                        'removed unsupported key `{}` with value `{}` from profile `{}`'.format(
                            internal_key, value, name
                        )
                    )
                    continue
            setattr(self, internal_key, value)

        # Create a default UUID if not specified
        if self.uuid is None:
            from uuid import uuid4
            self.uuid = uuid4().hex

        # Currently, whether a profile is a test profile is solely determined by its name starting with 'test_'
        self._test_profile = bool(self.name.startswith('test_'))

    @property
    def uuid(self):
        """Return the profile uuid.

        :return: string UUID
        """
        try:
            return self._attributes[self.KEY_UUID]
        except KeyError:
            return None

    @uuid.setter
    def uuid(self, value):
        self._attributes[self.KEY_UUID] = value

    @property
    def default_user(self):
        return self._attributes[self.KEY_DEFAULT_USER]

    @default_user.setter
    def default_user(self, value):
        self._attributes[self.KEY_DEFAULT_USER] = value

    @property
    def database_engine(self):
        return self._attributes[self.KEY_DATABASE_ENGINE]

    @database_engine.setter
    def database_engine(self, value):
        self._attributes[self.KEY_DATABASE_ENGINE] = value

    @property
    def database_backend(self):
        return self._attributes[self.KEY_DATABASE_BACKEND]

    @database_backend.setter
    def database_backend(self, value):
        self._attributes[self.KEY_DATABASE_BACKEND] = value

    @property
    def database_name(self):
        return self._attributes[self.KEY_DATABASE_NAME]

    @database_name.setter
    def database_name(self, value):
        self._attributes[self.KEY_DATABASE_NAME] = value

    @property
    def database_port(self):
        return self._attributes[self.KEY_DATABASE_PORT]

    @database_port.setter
    def database_port(self, value):
        self._attributes[self.KEY_DATABASE_PORT] = value

    @property
    def database_hostname(self):
        return self._attributes[self.KEY_DATABASE_HOSTNAME]

    @database_hostname.setter
    def database_hostname(self, value):
        self._attributes[self.KEY_DATABASE_HOSTNAME] = value

    @property
    def database_username(self):
        return self._attributes[self.KEY_DATABASE_USERNAME]

    @database_username.setter
    def database_username(self, value):
        self._attributes[self.KEY_DATABASE_USERNAME] = value

    @property
    def database_password(self):
        return self._attributes[self.KEY_DATABASE_PASSWORD]

    @database_password.setter
    def database_password(self, value):
        self._attributes[self.KEY_DATABASE_PASSWORD] = value

    @property
    def repository_uri(self):
        return self._attributes[self.KEY_REPOSITORY_URI]

    @repository_uri.setter
    def repository_uri(self, value):
        self._attributes[self.KEY_REPOSITORY_URI] = value

    @property
    def options(self):
        self._attributes.setdefault(self.KEY_OPTIONS, {})
        return self._attributes[self.KEY_OPTIONS]

    @options.setter
    def options(self, value):
        self._attributes[self.KEY_OPTIONS] = value

    def get_option(self, option_key, default=None):
        return self.options.get(option_key, default)

    def set_option(self, option_key, value, override=True):
        """Set a configuration option for a certain scope.

        :param option_key: the key of the configuration option
        :param option_value: the option value
        :param override: boolean, if False, will not override the option if it already exists
        """
        if option_key not in self.options or override:
            self.options[option_key] = value

    def unset_option(self, option_key):
        self.options.pop(option_key, None)

    @property
    def name(self):
        """Return the profile name.

        :return: the profile name
        """
        return self._name

    @property
    def dictionary(self):
        """Return the profile attributes as a dictionary with keys as it is stored in the config

        :return: the profile configuration dictionary
        """
        return self._attributes

    @property
    def rmq_prefix(self):
        """Return the prefix that should be used for RMQ resources

        :return: the rmq prefix string
        """
        return self.RMQ_PREFIX.format(uuid=self.uuid)

    @property
    def is_test_profile(self):
        """Return whether the profile is a test profile

        :return: boolean, True if test profile, False otherwise
        """
        return self._test_profile

    @property
    def repository_path(self):
        """Return the absolute path of the repository configured for this profile.

        :return: absolute filepath of the profile's file repository
        """
        return self._parse_repository_uri()[1]

    def _parse_repository_uri(self):
        """
        This function validates the REPOSITORY_URI, that should be in the format protocol://address

        :note: At the moment, only the file protocol is supported.

        :return: a tuple (protocol, address).
        """
        import uritools
        parts = uritools.urisplit(self.repository_uri)

        if parts.scheme != u'file':
            raise exceptions.ConfigurationError('invalid repository protocol, only the local `file://` is supported')

        if not os.path.isabs(parts.path):
            raise exceptions.ConfigurationError('invalid repository URI: the path has to be absolute')

        return parts.scheme, os.path.expanduser(parts.path)

    def configure_repository(self):
        """Validates the configured repository and in the case of a file system repo makes sure the folder exists."""
        import errno

        try:
            os.makedirs(self.repository_path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise exceptions.ConfigurationError(
                    'could not create the configured repository `{}`: {}'.format(self.repository_path, str(exception))
                )

    @property
    def filepaths(self):
        """Return the filepaths used by this profile.

        :return: a dictionary of filepaths
        """
        return {
            'circus': {
                'log': CIRCUS_LOG_FILE_TEMPLATE.format(self.name),
                'pid': CIRCUS_PID_FILE_TEMPLATE.format(self.name),
                'port': CIRCUS_PORT_FILE_TEMPLATE.format(self.name),
                'socket': {
                    'file': CIRCUS_SOCKET_FILE_TEMPATE.format(self.name),
                    'controller': CIRCUS_CONTROLLER_SOCKET_TEMPLATE,
                    'pubsub': CIRCUS_PUBSUB_SOCKET_TEMPLATE,
                    'stats': CIRCUS_STATS_SOCKET_TEMPLATE,
                }
            },
            'daemon': {
                'log': DAEMON_LOG_FILE_TEMPLATE.format(self.name),
                'pid': DAEMON_PID_FILE_TEMPLATE.format(self.name),
            }
        }



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=no-name-in-module,no-member,import-error,cyclic-import
"""Utility functions specific to the Django backend."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import django

from aiida.backends.utils import validate_attribute_key, SettingsManager, Setting
from aiida.common import NotExistent

SCHEMA_VERSION_DB_KEY = 'db|schemaversion'
SCHEMA_VERSION_DB_DESCRIPTION = 'The version of the schema used in this database.'


class DjangoSettingsManager(SettingsManager):
    """Class to get, set and delete settings from the `DbSettings` table."""

    table_name = 'db_dbsetting'

    def validate_table_existence(self):
        """Verify that the `DbSetting` table actually exists.

        :raises: `~aiida.common.exceptions.NotExistent` if the settings table does not exist
        """
        from django.db import connection
        if self.table_name not in connection.introspection.table_names():
            raise NotExistent('the settings table does not exist')

    def get(self, key):
        """Return the setting with the given key.

        :param key: the key identifying the setting
        :return: Setting
        :raises: `~aiida.common.exceptions.NotExistent` if the settings does not exist
        """
        from aiida.backends.djsite.db.models import DbSetting

        self.validate_table_existence()
        setting = DbSetting.objects.filter(key=key).first()

        if setting is None:
            raise NotExistent('setting `{}` does not exist'.format(key))

        return Setting(setting.key, setting.val, setting.description, setting.time)

    def set(self, key, value, description=None):
        """Return the settings with the given key.

        :param key: the key identifying the setting
        :param value: the value for the setting
        :param description: optional setting description
        """
        from aiida.backends.djsite.db.models import DbSetting

        self.validate_table_existence()
        validate_attribute_key(key)

        other_attribs = dict()
        if description is not None:
            other_attribs['description'] = description

        DbSetting.set_value(key, value, other_attribs=other_attribs)

    def delete(self, key):
        """Delete the setting with the given key.

        :param key: the key identifying the setting
        :raises: `~aiida.common.exceptions.NotExistent` if the settings does not exist
        """
        from aiida.backends.djsite.db.models import DbSetting

        self.validate_table_existence()

        try:
            DbSetting.del_value(key=key)
        except KeyError:
            raise NotExistent('setting `{}` does not exist'.format(key))


def load_dbenv(profile):
    """Load the database environment and ensure that the code and database schema versions are compatible.

    :param profile: instance of `Profile` whose database to load
    """
    _load_dbenv_noschemacheck(profile)
    check_schema_version(profile_name=profile.name)


def _load_dbenv_noschemacheck(profile):  # pylint: disable=unused-argument
    """Load the database environment without checking that code and database schema versions are compatible.

    This should ONLY be used internally, inside load_dbenv, and for schema migrations. DO NOT USE OTHERWISE!

    :param profile: instance of `Profile` whose database to load
    """
    os.environ['DJANGO_SETTINGS_MODULE'] = 'aiida.backends.djsite.settings'
    django.setup()


def unload_dbenv():
    """Unload the database environment.

    This means that the settings in `aiida.backends.djsite.settings` are "unset".
    This needs to implemented and will address #2813
    """


_aiida_autouser_cache = None  # pylint: disable=invalid-name


def migrate_database():
    """Migrate the database to the latest schema version."""
    from django.core.management import call_command
    call_command('migrate')


def check_schema_version(profile_name):
    """
    Check if the version stored in the database is the same of the version
    of the code.

    :note: if the DbSetting table does not exist, this function does not
      fail. The reason is to avoid to have problems before running the first
      migrate call.

    :note: if no version is found, the version is set to the version of the
      code. This is useful to have the code automatically set the DB version
      at the first code execution.

    :raise aiida.common.ConfigurationError: if the two schema versions do not match.
      Otherwise, just return.
    """
    # pylint: disable=duplicate-string-formatting-argument
    from django.db import connection

    import aiida.backends.djsite.db.models
    from aiida.common.exceptions import ConfigurationError

    # Do not do anything if the table does not exist yet
    if 'db_dbsetting' not in connection.introspection.table_names():
        return

    code_schema_version = aiida.backends.djsite.db.models.SCHEMA_VERSION
    db_schema_version = get_db_schema_version()

    if db_schema_version is None:
        # No code schema defined yet, I set it to the code version
        set_db_schema_version(code_schema_version)
        db_schema_version = get_db_schema_version()

    if code_schema_version != db_schema_version:
        raise ConfigurationError(
            'Database schema version {} is outdated compared to the code schema version {}\n'
            'Before you upgrade, make sure all calculations and workflows have finished running.\n'
            'If this is not the case, revert the code to the previous version and finish them first.\n'
            'To migrate the database to the current version, run the following commands:'
            '\n  verdi -p {} daemon stop\n  verdi -p {} database migrate'.format(
                db_schema_version, code_schema_version, profile_name, profile_name
            )
        )


def set_db_schema_version(version):
    """
    Set the schema version stored in the DB. Use only if you know what you are doing.
    """
    from aiida.backends.utils import get_settings_manager
    manager = get_settings_manager()
    return manager.set(SCHEMA_VERSION_DB_KEY, version, description=SCHEMA_VERSION_DB_DESCRIPTION)


def get_db_schema_version():
    """
    Get the current schema version stored in the DB. Return None if
    it is not stored.
    """
    from django.db.utils import ProgrammingError
    from aiida.manage.manager import get_manager
    backend = get_manager()._load_backend(schema_check=False)  # pylint: disable=protected-access

    try:
        result = backend.execute_raw(r"""SELECT tval FROM db_dbsetting WHERE key = 'db|schemaversion';""")
    except ProgrammingError:
        result = backend.execute_raw(r"""SELECT val FROM db_dbsetting WHERE key = 'db|schemaversion';""")

    return result[0][0]


def delete_nodes_and_connections_django(pks_to_delete):  # pylint: disable=invalid-name
    """
    Delete all nodes corresponding to pks in the input.
    :param pks_to_delete: A list, tuple or set of pks that should be deleted.
    """
    from django.db import transaction
    from django.db.models import Q
    from aiida.backends.djsite.db import models
    with transaction.atomic():
        # This is fixed in pylint-django>=2, but this supports only py3
        # pylint: disable=no-member
        # Delete all links pointing to or from a given node
        models.DbLink.objects.filter(Q(input__in=pks_to_delete) | Q(output__in=pks_to_delete)).delete()
        # now delete nodes
        models.DbNode.objects.filter(pk__in=pks_to_delete).delete()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

"""
In this file various data management functions, needed for the SQLA test,
are added. They are "heavily inspired" by the sqlalchemy_utils.functions.database
(SQLAlchemy-Utils package).

However, they were corrected to work properly with a SQlAlchemy and PostgreSQL.
The main problem of the SQLAlchemy-Utils that were rewritten was that they
were not properly disposing the (SQLA) engine, resulting to error messages
from PostgreSQL.

"""


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from sqlalchemy_utils.functions.database import drop_database


def new_database(uri):
    """Drop the database at ``uri`` and create a brand new one. """
    destroy_database(uri)
    create_database(uri)


def destroy_database(uri):
    """Destroy the database at ``uri``, if it exists. """
    if database_exists(uri):
        drop_database(uri)


def database_exists(url):
    """Check if a database exists.

    This is a modification of sqlalchemy_utils.functions.database.database_exists
    since the latter one did not correctly work with SQLAlchemy and PostgreSQL.

    :param url: A SQLAlchemy engine URL.

    Performs backend-specific testing to quickly determine if a database
    exists on the server.

    """
    from sqlalchemy.engine.url import make_url
    from copy import copy
    import sqlalchemy as sa

    url = copy(make_url(url))
    database = url.database
    if url.drivername.startswith('postgresql'):
        url.database = 'template1'
    else:
        url.database = None

    engine = sa.create_engine(url)

    try:
        if engine.dialect.name == 'postgresql':
            text = "SELECT 1 FROM pg_database WHERE datname='%s'" % database
            return bool(engine.execute(text).scalar())

        else:
            raise Exception('Only PostgreSQL is supported.')
    finally:
        engine.dispose()


def create_database(url, encoding='utf8'):
    """Issue the appropriate CREATE DATABASE statement.

    This is a modification of sqlalchemy_utils.functions.database.create_database
    since the latter one did not correctly work with SQLAlchemy and PostgreSQL.

    :param url: A SQLAlchemy engine URL.
    :param encoding: The encoding to create the database as.


    It currently supports only PostgreSQL and the psycopg2 driver.
    """

    from sqlalchemy.engine.url import make_url
    from sqlalchemy_utils.functions.orm import quote
    from copy import copy
    import sqlalchemy as sa

    url = copy(make_url(url))

    database = url.database

    # A default PostgreSQL database to connect
    url.database = 'template1'

    engine = sa.create_engine(url)

    try:
        if engine.dialect.name == 'postgresql' and engine.driver == 'psycopg2':
            from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
            engine.raw_connection().set_isolation_level(
                ISOLATION_LEVEL_AUTOCOMMIT
            )

            text = "CREATE DATABASE {0} ENCODING '{1}'".format(
                quote(engine, database),
                encoding
            )

            engine.execute(text)

        else:
            raise Exception('Only PostgreSQL with the psycopg2 driver is '
                            'supported.')
    finally:
        engine.dispose()


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This module provides list functionality to all data types.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.cmdline.params import options

LIST_OPTIONS = [
    options.GROUPS,
    options.PAST_DAYS,
    options.ALL_USERS,
    options.RAW,
]


def list_options(func):
    """Creates a decorator with all the options."""
    for option in reversed(LIST_OPTIONS):
        func = option()(func)

    return func


def query(datatype, project, past_days, group_pks, all_users):
    """
    Perform the query
    """
    import datetime

    from aiida import orm
    from aiida.common import timezone

    qbl = orm.QueryBuilder()
    if all_users is False:
        user = orm.User.objects.get_default()
        qbl.append(orm.User, tag='creator', filters={'email': user.email})
    else:
        qbl.append(orm.User, tag='creator')

    # If there is a time restriction
    data_filters = {}
    if past_days is not None:
        now = timezone.now()
        n_days_ago = now - datetime.timedelta(days=past_days)
        data_filters.update({'ctime': {'>=': n_days_ago}})

    qbl.append(datatype, tag='data', with_user='creator', filters=data_filters, project=project)

    # If there is a group restriction
    if group_pks is not None:
        group_filters = dict()
        group_filters.update({'id': {'in': group_pks}})
        qbl.append(orm.Group, tag='group', filters=group_filters, with_node='data')

    qbl.order_by({datatype: {'ctime': 'asc'}})

    object_list = qbl.distinct()
    return object_list.all()


# pylint: disable=unused-argument,too-many-arguments
def data_list(datatype, columns, elements, elements_only, formula_mode, past_days, groups, all_users):
    """
    List stored objects
    """
    columns_dict = {
        'ID': 'id',
        'Id': 'id',
        'Ctime': 'ctime',
        'Label': 'label',
        'Formula': 'attributes.formula',
        'Kinds': 'attributes.kinds',
        'Sites': 'attributes.sites',
        'Formulae': 'attributes.formulae',
        'Source': 'attributes.source',
        'Source.URI': 'attributes.source.uri',
    }
    project = [columns_dict[k] for k in columns]
    group_pks = None
    if groups is not None:
        group_pks = [g.pk for g in groups]
    return query(datatype, project, past_days, group_pks, all_users)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the classes in `aiida.engine.processes.calcjobs.manager`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import time

import tornado

from aiida.orm import AuthInfo, User
from aiida.backends.testbase import AiidaTestCase
from aiida.engine.processes.calcjobs.manager import JobManager, JobsList
from aiida.engine.transports import TransportQueue


class TestJobManager(AiidaTestCase):
    """Test the `aiida.engine.processes.calcjobs.manager.JobManager` class."""

    def setUp(self):
        super(TestJobManager, self).setUp()
        self.loop = tornado.ioloop.IOLoop()
        self.transport_queue = TransportQueue(self.loop)
        self.user = User.objects.get_default()
        self.auth_info = AuthInfo(self.computer, self.user).store()
        self.manager = JobManager(self.transport_queue)

    def tearDown(self):
        super(TestJobManager, self).tearDown()
        AuthInfo.objects.delete(self.auth_info.pk)

    def test_get_jobs_list(self):
        """Test the `JobManager.get_jobs_list` method."""
        jobs_list = self.manager.get_jobs_list(self.auth_info)
        self.assertIsInstance(jobs_list, JobsList)

        # Calling the method again, should return the exact same instance of `JobsList`
        self.assertEqual(self.manager.get_jobs_list(self.auth_info), jobs_list)

    def test_request_job_info_update(self):
        """Test the `JobManager.request_job_info_update` method."""
        with self.manager.request_job_info_update(self.auth_info, job_id=1) as request:
            self.assertIsInstance(request, tornado.concurrent.Future)


class TestJobsList(AiidaTestCase):
    """Test the `aiida.engine.processes.calcjobs.manager.JobsList` class."""

    def setUp(self):
        super(TestJobsList, self).setUp()
        self.loop = tornado.ioloop.IOLoop()
        self.transport_queue = TransportQueue(self.loop)
        self.user = User.objects.get_default()
        self.auth_info = AuthInfo(self.computer, self.user).store()
        self.jobs_list = JobsList(self.auth_info, self.transport_queue)

    def tearDown(self):
        super(TestJobsList, self).tearDown()
        AuthInfo.objects.delete(self.auth_info.pk)

    def test_get_minimum_update_interval(self):
        """Test the `JobsList.get_minimum_update_interval` method."""
        minimum_poll_interval = self.auth_info.computer.get_minimum_job_poll_interval()
        self.assertEqual(self.jobs_list.get_minimum_update_interval(), minimum_poll_interval)

    def test_last_updated(self):
        """Test the `JobsList.last_updated` method."""
        jobs_list = JobsList(self.auth_info, self.transport_queue)
        self.assertEqual(jobs_list.last_updated, None)

        last_updated = time.time()
        jobs_list = JobsList(self.auth_info, self.transport_queue, last_updated=last_updated)
        self.assertEqual(jobs_list.last_updated, last_updated)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import collections
import six

from aiida.backends import BACKEND_SQLA, BACKEND_DJANGO
from aiida.common.exceptions import ConfigurationError, ValidationError
from aiida.manage import configuration

AIIDA_ATTRIBUTE_SEP = '.'


Setting = collections.namedtuple('Setting', ['key', 'value', 'description', 'time'])


class SettingsManager(object):
    """Class to get, set and delete settings from the `DbSettings` table."""

    @abc.abstractmethod
    def get(self, key):
        """Return the setting with the given key.

        :param key: the key identifying the setting
        :return: Setting
        :raises: `~aiida.common.exceptions.NotExistent` if the settings does not exist
        """

    @abc.abstractmethod
    def set(self, key, value, description=None):
        """Return the settings with the given key.

        :param key: the key identifying the setting
        :param value: the value for the setting
        :param description: optional setting description
        """

    @abc.abstractmethod
    def delete(self, key):
        """Delete the setting with the given key.

        :param key: the key identifying the setting
        :raises: `~aiida.common.exceptions.NotExistent` if the settings does not exist
        """


def get_settings_manager():
    if configuration.PROFILE.database_backend == BACKEND_DJANGO:
        from aiida.backends.djsite.utils import DjangoSettingsManager
        manager = DjangoSettingsManager()
    elif configuration.PROFILE.database_backend == BACKEND_SQLA:
        from aiida.backends.sqlalchemy.utils import SqlaSettingsManager
        manager = SqlaSettingsManager()
    else:
        raise Exception('unknown backend type `{}`'.format(configuration.PROFILE.database_backend))

    return manager


def validate_attribute_key(key):
    """
    Validate the key string to check if it is valid (e.g., if it does not
    contain the separator symbol.).

    :return: None if the key is valid
    :raise aiida.common.ValidationError: if the key is not valid
    """
    if not isinstance(key, six.string_types):
        raise ValidationError('The key must be a string.')
    if not key:
        raise ValidationError('The key cannot be an empty string.')
    if AIIDA_ATTRIBUTE_SEP in key:
        raise ValidationError("The separator symbol '{}' cannot be present "
                              'in the key of attributes, extras, etc.'.format(AIIDA_ATTRIBUTE_SEP))


def load_dbenv(profile=None, *args, **kwargs):
    if configuration.PROFILE.database_backend == BACKEND_SQLA:
        # Maybe schema version should be also checked for SQLAlchemy version.
        from aiida.backends.sqlalchemy.utils \
            import load_dbenv as load_dbenv_sqlalchemy
        to_return = load_dbenv_sqlalchemy(profile=profile, *args, **kwargs)
    elif configuration.PROFILE.database_backend == BACKEND_DJANGO:
        from aiida.backends.djsite.utils import load_dbenv as load_dbenv_django
        to_return = load_dbenv_django(profile=profile, *args, **kwargs)
    else:
        raise ConfigurationError('Invalid configuration.PROFILE.database_backend: {}'.format(
            configuration.PROFILE.database_backend))

    return to_return


def _load_dbenv_noschemacheck(profile=None, *args, **kwargs):
    if configuration.PROFILE.database_backend == BACKEND_SQLA:
        # Maybe schema version should be also checked for SQLAlchemy version.
        from aiida.backends.sqlalchemy.utils \
            import _load_dbenv_noschemacheck as _load_dbenv_noschemacheck_sqlalchemy
        to_return = _load_dbenv_noschemacheck_sqlalchemy(profile=profile, *args, **kwargs)
    elif configuration.PROFILE.database_backend == BACKEND_DJANGO:
        from aiida.backends.djsite.utils import _load_dbenv_noschemacheck as _load_dbenv_noschemacheck_django
        to_return = _load_dbenv_noschemacheck_django(profile=profile, *args, **kwargs)
    else:
        raise ConfigurationError('Invalid configuration.PROFILE.database_backend: {}'.format(configuration.PROFILE.database_backend))

    return to_return


def delete_nodes_and_connections(pks):
    if configuration.PROFILE.database_backend == BACKEND_DJANGO:
        from aiida.backends.djsite.utils import delete_nodes_and_connections_django as delete_nodes_backend
    elif configuration.PROFILE.database_backend == BACKEND_SQLA:
        from aiida.backends.sqlalchemy.utils import delete_nodes_and_connections_sqla as delete_nodes_backend
    else:
        raise Exception('unknown backend {}'.format(configuration.PROFILE.database_backend))

    delete_nodes_backend(pks)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test export file migration from export version 0.6 to 0.7"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils.archives import get_json_files
from aiida.tools.importexport.migration.utils import verify_metadata_version
from aiida.tools.importexport.migration.v06_to_v07 import (
    migrate_v6_to_v7, migration_data_migration_legacy_process_attributes
)


class TestMigrateV06toV07(AiidaTestCase):
    """Test migration of export files from export version 0.6 to 0.7"""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestMigrateV06toV07, cls).setUpClass(*args, **kwargs)

        # Utility helpers
        cls.external_archive = {'filepath': 'archives', 'external_module': 'aiida-export-migration-tests'}
        cls.core_archive = {'filepath': 'export/migrate'}

    def test_migrate_v6_to_v7(self):
        """Test migration for file containing complete v0.6 era possibilities"""
        from aiida import get_version

        # Get metadata.json and data.json as dicts from v0.6 file archive
        metadata_v6, data_v6 = get_json_files('export_v0.6_simple.aiida', **self.core_archive)
        verify_metadata_version(metadata_v6, version='0.6')

        # Get metadata.json and data.json as dicts from v0.7 file archive
        metadata_v7, data_v7 = get_json_files('export_v0.7_simple.aiida', **self.core_archive)
        verify_metadata_version(metadata_v7, version='0.7')

        # Migrate to v0.7
        migrate_v6_to_v7(metadata_v6, data_v6)
        verify_metadata_version(metadata_v6, version='0.7')

        # Remove AiiDA version, since this may change irregardless of the migration function
        metadata_v6.pop('aiida_version')
        metadata_v7.pop('aiida_version')

        # Assert conversion message in `metadata.json` is correct and then remove it for later assertions
        self.maxDiff = None  # pylint: disable=invalid-name
        conversion_message = 'Converted from version 0.6 to 0.7 with AiiDA v{}'.format(get_version())
        self.assertEqual(
            metadata_v6.pop('conversion_info')[-1],
            conversion_message,
            msg='The conversion message after migration is wrong'
        )
        metadata_v7.pop('conversion_info')

        # Assert changes were performed correctly
        self.assertDictEqual(
            metadata_v6,
            metadata_v7,
            msg='After migration, metadata.json should equal intended metadata.json from archives'
        )
        self.assertDictEqual(
            data_v6, data_v7, msg='After migration, data.json should equal intended data.json from archives'
        )

    def test_migrate_v6_to_v7_complete(self):
        """Test migration for file containing complete v0.6 era possibilities"""
        # Get metadata.json and data.json as dicts from v0.6 file archive
        metadata, data = get_json_files('export_v0.6_manual.aiida', **self.external_archive)
        verify_metadata_version(metadata, version='0.6')

        # Migrate to v0.7
        migrate_v6_to_v7(metadata, data)
        verify_metadata_version(metadata, version='0.7')

        self.maxDiff = None  # pylint: disable=invalid-name
        # Check attributes of process.* nodes
        illegal_attrs = {'_sealed', '_finished', '_failed', '_aborted', '_do_abort'}
        new_attrs = {'sealed': True}
        for node_pk, attrs in data['node_attributes'].items():
            if data['export_data']['Node'][node_pk]['node_type'].startswith('process.'):
                # Check if illegal attributes were removed successfully
                for attr in illegal_attrs:
                    self.assertNotIn(
                        attr,
                        attrs,
                        msg="key '{}' should have been removed from attributes for Node <pk={}>".format(attr, node_pk)
                    )

                # Check new attributes were added successfully
                for attr in new_attrs:
                    self.assertIn(
                        attr, attrs, msg="key '{}' was not added to attributes for Node <pk={}>".format(attr, node_pk)
                    )
                    self.assertEqual(
                        attrs[attr],
                        new_attrs[attr],
                        msg="key '{}' should have had the value {}, but did instead have {}".format(
                            attr, new_attrs[attr], attrs[attr]
                        )
                    )

        # Check Attribute and Link have been removed
        illegal_entities = {'Attribute', 'Link'}
        for dict_ in {'unique_identifiers', 'all_fields_info'}:
            for entity in illegal_entities:
                self.assertNotIn(
                    entity,
                    metadata[dict_],
                    msg="key '{}' should have been removed from '{}' in metadata.json".format(entity, dict_)
                )

    def test_migration_0040_corrupt_archive(self):
        """Check CorruptArchive is raised for different cases during migration 0040"""
        from aiida.tools.importexport.common.exceptions import CorruptArchive

        # data has one "valid" entry, in the form of Node <PK=42>.
        # At least it has the needed key `node_type`.
        # data also has one "invalid" entry, in form of Node <PK=25>.
        # It is listed in `node_attributes`, but not in `export_data.Node`
        # Also Node <PK=25> is present in `export_data.Node`, but not in `node_attributes`.
        # This is not fine, but will not (and should not) be caught by this particular migration function
        data = {
            'node_attributes': {
                25: {},
                42: {}
            },
            'export_data': {
                'Node': {
                    42: {
                        'node_type': 'data.int.Int.'
                    },
                    52: {
                        'node_type': 'data.dict.Dict.'
                    }
                }
            }
        }

        with self.assertRaises(CorruptArchive) as exc:
            migration_data_migration_legacy_process_attributes(data)

        self.assertIn('Your export archive is corrupt! Org. exception:', str(exc.exception))

        # data has one "valid" entry, in the form of Node <PK=52>.
        # data also has one "invalid" entry, in form of Node <PK=42>.
        # It is in an active process state ('waiting')
        data = {
            'node_attributes': {
                52: {
                    'process_state': 'finished',
                    '_sealed': True
                },
                42: {
                    'process_state': 'waiting'
                }
            },
            'export_data': {
                'Node': {
                    42: {
                        'node_type': 'process.calculation.calcjob.CalcJobNode.'
                    },
                    52: {
                        'node_type': 'process.calculation.calcjob.CalcJobNode.'
                    }
                }
            }
        }

        with self.assertRaises(CorruptArchive) as exc:
            migration_data_migration_legacy_process_attributes(data)

        self.assertIn('Your export archive is corrupt! Please see the log-file', str(exc.exception))

    def test_migration_0040_no_process_state(self):
        """Check old ProcessNodes without a `process_state` can be migrated"""
        # data has one "new" entry, in the form of Node <PK=52>.
        # data also has one "old" entry, in form of Node <PK=42>.
        # It doesn't have a `process_state` attribute (nor a `sealed` or `_sealed`)
        data = {
            'node_attributes': {
                52: {
                    'process_state': 'finished',
                    'sealed': True
                },
                42: {}
            },
            'export_data': {
                'Node': {
                    42: {
                        'node_type': 'process.workflow.workfunction.WorkFunctionNode.'
                    },
                    52: {
                        'node_type': 'process.calculation.calcjob.CalcJobNode.'
                    }
                }
            }
        }

        migration_data_migration_legacy_process_attributes(data)

        self.assertDictEqual(data['node_attributes'][42], {'sealed': True})



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""A transport queue to batch process multiple tasks that require a Transport."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from collections import namedtuple
import contextlib
import logging
import traceback
from tornado import concurrent, gen, ioloop

_LOGGER = logging.getLogger(__name__)


class TransportRequest(object):
    """ Information kept about request for a transport object """

    # pylint: disable=too-few-public-methods
    def __init__(self):
        super(TransportRequest, self).__init__()
        self.future = concurrent.Future()
        self.count = 0


class TransportQueue(object):
    """
    A queue to get transport objects from authinfo.  This class allows clients
    to register their interest in a transport object which will be provided at
    some point in the future.

    Internally the class will wait for a specific interval at the end of which
    it will open the transport and give it to all the clients that asked for it
    up to that point.  This way opening of transports (a costly operation) can
    be minimised.
    """
    AuthInfoEntry = namedtuple('AuthInfoEntry', ['authinfo', 'transport', 'callbacks', 'callback_handle'])

    def __init__(self, loop=None):
        """
        :param loop: The event loop to use, will use `tornado.ioloop.IOLoop.current()` if not supplied
        :type loop: :class:`tornado.ioloop.IOLoop`
        """
        self._loop = loop if loop is not None else ioloop.IOLoop.current()
        self._transport_requests = {}

    def loop(self):
        """ Get the loop being used by this transport queue """
        return self._loop

    @contextlib.contextmanager
    def request_transport(self, authinfo):
        """
        Request a transport from an authinfo.  Because the client is not allowed to
        request a transport immediately they will instead be given back a future
        that can be yielded to get the transport::

            @tornado.gen.coroutine
            def transport_task(transport_queue, authinfo):
                with transport_queue.request_transport(authinfo) as request:
                    transport = yield request
                    # Do some work with the transport

        :param authinfo: The authinfo to be used to get transport
        :return: A future that can be yielded to give the transport
        """
        open_callback_handle = None
        transport_request = self._transport_requests.get(authinfo.id, None)

        if transport_request is None:
            # There is no existing request for this transport (i.e. on this authinfo)
            transport_request = TransportRequest()
            self._transport_requests[authinfo.id] = transport_request

            transport = authinfo.get_transport()
            safe_open_interval = transport.get_safe_open_interval()

            def do_open():
                """ Actually open the transport """
                if transport_request.count > 0:
                    # The user still wants the transport so open it
                    _LOGGER.debug('Transport request opening transport for %s', authinfo)
                    try:
                        transport.open()
                    except Exception as exception:  # pylint: disable=broad-except
                        _LOGGER.error('exception occurred while trying to open transport:\n %s', exception)
                        transport_request.future.set_exception(exception)

                        # Cleanup of the stale TransportRequest with the excepted transport future
                        self._transport_requests.pop(authinfo.id, None)
                    else:
                        transport_request.future.set_result(transport)

            # Save the handle so that we can cancel the callback if the user no longer wants it
            open_callback_handle = self._loop.call_later(safe_open_interval, do_open)

        try:
            transport_request.count += 1
            yield transport_request.future
        except gen.Return:
            # Have to have this special case so tornado returns are propagated up to the loop
            raise
        except Exception:
            _LOGGER.error('Exception whilst using transport:\n%s', traceback.format_exc())
            raise
        finally:
            transport_request.count -= 1
            assert transport_request.count >= 0, 'Transport request count dropped below 0!'
            # Check if there are no longer any users that want the transport
            if transport_request.count == 0:
                if transport_request.future.done():
                    _LOGGER.debug('Transport request closing transport for %s', authinfo)
                    transport_request.future.result().close()
                elif open_callback_handle is not None:
                    self._loop.remove_timeout(open_callback_handle)

                self._transport_requests.pop(authinfo.id, None)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import six

from aiida.tools.dbimporters.baseclasses import (DbImporter, DbSearchResults,
                                                 UpfEntry)


class NnincDbImporter(DbImporter):
    """
    Database importer for NNIN/C Pseudopotential Virtual Vault.
    """

    def _str_clause(self, key, alias, values):
        """
        Returns part of HTTP GET query for querying string fields.
        """
        if not isinstance(values, six.string_types):
            raise ValueError("incorrect value for keyword '{}' -- only "
                             'strings and integers are accepted'.format(alias))
        return '{}={}'.format(key, values)

    _keywords = {'xc_approximation': ['frmxcprox', _str_clause],
                 'xc_type': ['frmxctype', _str_clause],
                 'pseudopotential_class': ['frmspclass', _str_clause],
                 'element': ['element', None]}

    def __init__(self, **kwargs):
        self._query_url = 'http://nninc.cnf.cornell.edu/dd_search.php'
        self.setup_db(**kwargs)

    def query_get(self, **kwargs):
        """
        Forms a HTTP GET query for querying the NNIN/C Pseudopotential
        Virtual Vault.

        :return: a string with HTTP GET statement.
        """
        get_parts = []
        for key in self._keywords.keys():
            if key in kwargs.keys():
                values = kwargs.pop(key)
                if self._keywords[key][1] is not None:
                    get_parts.append(
                        self._keywords[key][1](self,
                                               self._keywords[key][0],
                                               key,
                                               values))

        if kwargs.keys():
            raise NotImplementedError("search keyword(s) '"
                                      "', '".join(kwargs.keys()) + \
                                      "' is(are) not implemented for NNIN/C")

        return self._query_url + '?' + '&'.join(get_parts)

    def query(self, **kwargs):
        """
        Performs a query on the NNIN/C Pseudopotential Virtual Vault using
        ``keyword = value`` pairs, specified in ``kwargs``.

        :return: an instance of
            :py:class:`aiida.tools.dbimporters.plugins.nninc.NnincSearchResults`.
        """
        from six.moves import urllib
        import re

        query = self.query_get(**kwargs)
        response = urllib.request.urlopen(query).read()
        results = re.findall("psp_files/([^']+)\.UPF", response)

        elements = kwargs.get('element', None)
        if elements and not isinstance(elements, list):
            elements = [elements]

        if elements:
            results_now = set()
            for psp in results:
                for element in elements:
                    if psp.startswith('{}.'.format(element)):
                        results_now = results_now | set([psp])
            results = list(results_now)

        return NnincSearchResults([{'id': x} for x in results])

    def setup_db(self, query_url=None, **kwargs):
        """
        Changes the database connection details.
        """
        if query_url:
            self._query_url = query_url

        if kwargs.keys():
            raise NotImplementedError( \
                "unknown database connection parameter(s): '" + \
                "', '".join(kwargs.keys()) + \
                "', available parameters: 'query_url'")

    def get_supported_keywords(self):
        """
        Returns the list of all supported query keywords.

        :return: list of strings
        """
        return self._keywords.keys()


class NnincSearchResults(DbSearchResults):
    """
    Results of the search, performed on NNIN/C Pseudopotential Virtual
    Vault.
    """
    _base_url = 'http://nninc.cnf.cornell.edu/psp_files/'

    def __init__(self, results):
        super(NnincSearchResults, self).__init__(results)
        self._return_class = NnincEntry

    def __len__(self):
        return len(self._results)

    def _get_source_dict(self, result_dict):
        """
        Returns a dictionary, which is passed as kwargs to the created
        DbEntry instance, describing the source of the entry.

        :param result_dict: dictionary, describing an entry in the results.
        """
        return {'id': result_dict['id']}

    def _get_url(self, result_dict):
        """
        Returns an URL of an entry CIF file.

        :param result_dict: dictionary, describing an entry in the results.
        """
        return self._base_url + result_dict['id'] + '.UPF'


class NnincEntry(UpfEntry):
    """
    Represents an entry from NNIN/C Pseudopotential Virtual Vault.
    """

    def __init__(self, uri, **kwargs):
        """
        Creates an instance of
        :py:class:`aiida.tools.dbimporters.plugins.nninc.NnincEntry`, related
        to the supplied URI.
        """
        super(NnincEntry, self).__init__(db_name='NNIN/C Pseudopotential Virtual Vault',
                                         db_uri='http://nninc.cnf.cornell.edu',
                                         uri=uri,
                                         **kwargs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

from django.db import models, migrations

from aiida.backends.djsite.db.migrations import upgrade_schema_version


REVISION = '1.0.2'
DOWN_REVISION = '1.0.1'


def fix_calc_states(apps, schema_editor):
    from aiida.orm.utils import load_node

    # These states should never exist in the database but we'll play it safe
    # and deal with them if they do
    DbCalcState = apps.get_model('db', 'DbCalcState')
    for calc_state in DbCalcState.objects.filter(
            state__in=[u'UNDETERMINED', u'NOTFOUND']):
        old_state = calc_state.state
        calc_state.state = u'FAILED'
        calc_state.save()
        # Now add a note in the log to say what we've done
        calc = load_node(pk=calc_state.dbnode.pk)
        calc.logger.warning(
            'Job state {} found for calculation {} which should never be in '
            'the database. Changed state to FAILED.'.format(
                old_state, calc_state.dbnode.pk))


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dbcalcstate',
            name='state',
            # The UNDETERMINED and NOTFOUND 'states' were removed as these
            # don't make sense
            field=models.CharField(db_index=True, max_length=25,
                                   choices=[(u'RETRIEVALFAILED', u'RETRIEVALFAILED'), (u'COMPUTED', u'COMPUTED'),
                                            (u'RETRIEVING', u'RETRIEVING'), (u'WITHSCHEDULER', u'WITHSCHEDULER'),
                                            (u'SUBMISSIONFAILED', u'SUBMISSIONFAILED'), (u'PARSING', u'PARSING'),
                                            (u'FAILED', u'FAILED'), (u'FINISHED', u'FINISHED'),
                                            (u'TOSUBMIT', u'TOSUBMIT'), (u'SUBMITTING', u'SUBMITTING'),
                                            (u'IMPORTED', u'IMPORTED'), (u'NEW', u'NEW'),
                                            (u'PARSINGFAILED', u'PARSINGFAILED')]),
            preserve_default=True,
        ),
        # Fix up any calculation states that had one of the removed states
        migrations.RunPython(fix_calc_states),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Unit tests for the BackendComment and BackendCommentCollection classes."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from datetime import datetime
from uuid import UUID

from six.moves import zip

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.common import timezone
from aiida.common import exceptions


class TestBackendComment(AiidaTestCase):
    """Test BackendComment."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestBackendComment, cls).setUpClass(*args, **kwargs)
        cls.computer = cls.computer.backend_entity  # Unwrap the `Computer` instance to `BackendComputer`
        cls.user = cls.backend.users.create(email='tester@localhost').store()

    def setUp(self):
        super(TestBackendComment, self).setUp()
        self.node = self.backend.nodes.create(
            node_type='', user=self.user, computer=self.computer, label='label', description='description'
        ).store()
        self.comment_content = 'comment content'

    def create_comment(self, **kwargs):
        """Create BackendComment"""
        node = kwargs['node'] if 'node' in kwargs else self.node
        user = kwargs['user'] if 'user' in kwargs else self.user
        ctime = kwargs['ctime'] if 'ctime' in kwargs else None
        mtime = kwargs['mtime'] if 'mtime' in kwargs else None

        return self.backend.comments.create(
            node=node, user=user, content=self.comment_content, ctime=ctime, mtime=mtime
        )

    def test_creation(self):
        """Test creation of a BackendComment and all its properties."""
        comment = self.backend.comments.create(node=self.node, user=self.user, content=self.comment_content)

        # Before storing
        self.assertIsNone(comment.id)
        self.assertIsNone(comment.pk)
        self.assertTrue(isinstance(comment.uuid, str))
        self.assertTrue(comment.node, self.node)
        self.assertTrue(isinstance(comment.ctime, datetime))
        self.assertIsNone(comment.mtime)
        self.assertTrue(comment.user, self.user)
        self.assertEqual(comment.content, self.comment_content)

        # Store the comment.ctime before the store as a reference
        now = timezone.now()
        comment_ctime_before_store = comment.ctime
        self.assertTrue(now > comment.ctime, '{} is not smaller than now {}'.format(comment.ctime, now))

        comment.store()
        comment_ctime = comment.ctime
        comment_mtime = comment.mtime

        # The comment.ctime should have been unchanged, but the comment.mtime should have changed
        self.assertEqual(comment.ctime, comment_ctime_before_store)
        self.assertIsNotNone(comment.mtime)
        self.assertTrue(now < comment.mtime, '{} is not larger than now {}'.format(comment.mtime, now))

        # After storing
        self.assertTrue(isinstance(comment.id, int))
        self.assertTrue(isinstance(comment.pk, int))
        self.assertTrue(isinstance(comment.uuid, str))
        self.assertTrue(comment.node, self.node)
        self.assertTrue(isinstance(comment.ctime, datetime))
        self.assertTrue(isinstance(comment.mtime, datetime))
        self.assertTrue(comment.user, self.user)
        self.assertEqual(comment.content, self.comment_content)

        # Try to construct a UUID from the UUID value to prove that it has a valid UUID
        UUID(comment.uuid)

        # Change a column, which should trigger the save, update the mtime but leave the ctime untouched
        comment.set_content('test')
        self.assertEqual(comment.ctime, comment_ctime)
        self.assertTrue(comment.mtime > comment_mtime)

    def test_creation_with_time(self):
        """
        Test creation of a BackendComment when passing the mtime and the ctime. The passed ctime and mtime
        should be respected since it is important for the correct import of nodes at the AiiDA import/export.
        """
        from aiida.tools.importexport.dbimport.backends.utils import deserialize_attributes

        ctime = deserialize_attributes('2019-02-27T16:20:12.245738', 'date')
        mtime = deserialize_attributes('2019-02-27T16:27:14.798838', 'date')

        comment = self.backend.comments.create(
            node=self.node, user=self.user, content=self.comment_content, mtime=mtime, ctime=ctime
        )

        # Check that the ctime and mtime are the given ones
        self.assertEqual(comment.ctime, ctime)
        self.assertEqual(comment.mtime, mtime)

        comment.store()

        # Check that the given values remain even after storing
        self.assertEqual(comment.ctime, ctime)
        self.assertEqual(comment.mtime, mtime)

    def test_delete(self):
        """Test `delete` method"""
        # Create Comment, making sure it exists
        comment = self.create_comment()
        comment.store()
        comment_uuid = str(comment.uuid)

        builder = orm.QueryBuilder().append(orm.Comment, project='uuid')
        no_of_comments = builder.count()
        found_comments_uuid = [_[0] for _ in builder.all()]
        self.assertIn(comment_uuid, found_comments_uuid)

        # Delete Comment, making sure it was deleted
        self.backend.comments.delete(comment.id)

        builder = orm.QueryBuilder().append(orm.Comment, project='uuid')
        self.assertEqual(builder.count(), no_of_comments - 1)
        found_comments_uuid = [_[0] for _ in builder.all()]
        self.assertNotIn(comment_uuid, found_comments_uuid)

    def test_delete_all(self):
        """Test `delete_all` method"""
        self.create_comment().store()
        self.assertGreater(len(orm.Comment.objects.all()), 0, msg='There should be Comments in the database')

        self.backend.comments.delete_all()
        self.assertEqual(len(orm.Comment.objects.all()), 0, msg='All Comments should have been deleted')

    def test_delete_many_no_filters(self):
        """Test `delete_many` method with empty filters"""
        self.create_comment().store()
        count = len(orm.Comment.objects.all())
        self.assertGreater(count, 0)

        # Pass empty filter to delete_many, making sure ValidationError is raised
        with self.assertRaises(exceptions.ValidationError):
            self.backend.comments.delete_many({})
        self.assertEqual(
            len(orm.Comment.objects.all()),
            count,
            msg='No Comments should have been deleted. There should still be {} Comment(s), '
            'however {} Comment(s) was/were found.'.format(count, len(orm.Comment.objects.all()))
        )

    def test_delete_many_ids(self):
        """Test `delete_many` method filtering on both `id` and `uuid`"""
        comment1 = self.create_comment()
        comment2 = self.create_comment()
        comment3 = self.create_comment()
        comment_uuids = []
        for comment in [comment1, comment2, comment3]:
            comment.store()
            comment_uuids.append(str(comment.uuid))

        # Make sure they exist
        count_comments_found = orm.QueryBuilder().append(orm.Comment, filters={'uuid': {'in': comment_uuids}}).count()
        self.assertEqual(
            count_comments_found,
            len(comment_uuids),
            msg='There should be {} Comments, instead {} Comment(s) was/were found'.format(
                len(comment_uuids), count_comments_found
            )
        )

        # Delete last two comments (comment2, comment3)
        filters = {'or': [{'id': comment2.id}, {'uuid': str(comment3.uuid)}]}
        self.backend.comments.delete_many(filters=filters)

        # Check they were deleted
        builder = orm.QueryBuilder().append(orm.Comment, filters={'uuid': {'in': comment_uuids}}, project='uuid').all()
        found_comments_uuid = [_[0] for _ in builder]
        self.assertEqual([comment_uuids[0]], found_comments_uuid)

    def test_delete_many_dbnode_id(self):
        """Test `delete_many` method filtering on `dbnode_id`"""
        # Create comments and separate node
        calc = self.backend.nodes.create(
            node_type='', user=self.user, computer=self.computer, label='label', description='description'
        ).store()
        comment1 = self.create_comment(node=calc)
        comment2 = self.create_comment()
        comment3 = self.create_comment()
        comment_uuids = []
        for comment in [comment1, comment2, comment3]:
            comment.store()
            comment_uuids.append(str(comment.uuid))

        # Make sure they exist
        count_comments_found = orm.QueryBuilder().append(orm.Comment, filters={'uuid': {'in': comment_uuids}}).count()
        self.assertEqual(
            count_comments_found,
            len(comment_uuids),
            msg='There should be {} Comments, instead {} Comment(s) was/were found'.format(
                len(comment_uuids), count_comments_found
            )
        )

        # Delete comments for self.node
        filters = {'dbnode_id': self.node.id}
        self.backend.comments.delete_many(filters=filters)

        # Check they were deleted
        builder = orm.QueryBuilder().append(orm.Comment, filters={'uuid': {'in': comment_uuids}}, project='uuid').all()
        found_comments_uuid = [_[0] for _ in builder]
        self.assertEqual([comment_uuids[0]], found_comments_uuid)

    # pylint: disable=too-many-locals
    def test_delete_many_ctime_mtime(self):
        """Test `delete_many` method filtering on `ctime` and `mtime`"""
        from datetime import timedelta

        # Initialization
        comment_uuids = []
        found_comments_ctime = []
        found_comments_mtime = []
        found_comments_uuid = []

        now = timezone.now()
        two_days_ago = now - timedelta(days=2)
        one_day_ago = now - timedelta(days=1)
        comment_times = [now, one_day_ago, two_days_ago]

        # Create comments
        comment1 = self.create_comment(ctime=now, mtime=now)
        comment2 = self.create_comment(ctime=one_day_ago, mtime=now)
        comment3 = self.create_comment(ctime=two_days_ago, mtime=one_day_ago)
        for comment in [comment1, comment2, comment3]:
            comment.store()
            comment_uuids.append(str(comment.uuid))

        # Make sure they exist with the correct times
        builder = orm.QueryBuilder().append(orm.Comment, project=['ctime', 'mtime', 'uuid'])
        self.assertGreater(builder.count(), 0)
        for comment in builder.all():
            found_comments_ctime.append(comment[0])
            found_comments_mtime.append(comment[1])
            found_comments_uuid.append(comment[2])
        for time, uuid in zip(comment_times, comment_uuids):
            self.assertIn(time, found_comments_ctime)
            self.assertIn(uuid, found_comments_uuid)
            if time != two_days_ago:
                self.assertIn(time, found_comments_mtime)

        # Delete comments that are created more than 1 hour ago,
        # unless they have been modified within 5 hours
        ctime_turning_point = now - timedelta(seconds=60 * 60)
        mtime_turning_point = now - timedelta(seconds=60 * 60 * 5)
        filters = {'and': [{'ctime': {'<': ctime_turning_point}}, {'mtime': {'<': mtime_turning_point}}]}
        self.backend.comments.delete_many(filters=filters)

        # Check only the most stale comment (comment3) was deleted
        builder = orm.QueryBuilder().append(orm.Comment, project='uuid')
        self.assertGreater(builder.count(), 1)  # There should still be at least 2
        found_comments_uuid = [_[0] for _ in builder.all()]
        self.assertNotIn(comment_uuids[2], found_comments_uuid)

        # Make sure the other comments were not deleted
        for comment_uuid in comment_uuids[:-1]:
            self.assertIn(comment_uuid, found_comments_uuid)

    def test_delete_many_user_id(self):
        """Test `delete_many` method filtering on `user_id`"""
        # Create comments and separate user
        user_two = self.backend.users.create(email='tester_two@localhost').store()
        comment1 = self.create_comment(user=user_two)
        comment2 = self.create_comment()
        comment3 = self.create_comment()
        comment_uuids = []
        for comment in [comment1, comment2, comment3]:
            comment.store()
            comment_uuids.append(str(comment.uuid))

        # Make sure they exist
        builder = orm.QueryBuilder().append(orm.Comment, project='uuid')
        self.assertGreater(builder.count(), 0)
        found_comments_uuid = [_[0] for _ in builder.all()]
        for comment_uuid in comment_uuids:
            self.assertIn(comment_uuid, found_comments_uuid)

        # Delete last comments for `self.user`
        filters = {'user_id': self.user.id}
        self.backend.comments.delete_many(filters=filters)

        # Check they were deleted
        builder = orm.QueryBuilder().append(orm.Comment, project='uuid')
        found_comments_uuid = [_[0] for _ in builder.all()]
        self.assertGreater(builder.count(), 0)
        for comment_uuid in comment_uuids[1:]:
            self.assertNotIn(comment_uuid, found_comments_uuid)

        # Make sure the first comment (comment1) was not deleted
        self.assertIn(comment_uuids[0], found_comments_uuid)

    def test_deleting_non_existent_entities(self):
        """Test deleting non-existent Comments for different cases"""
        comment = self.create_comment()
        comment.store()
        comment_id = comment.id
        comment_uuid = comment.uuid

        # Get a non-existent Comment
        valid_comment_found = True
        id_ = 0
        while valid_comment_found:
            id_ += 1
            builder = orm.QueryBuilder().append(orm.Comment, filters={'id': id_})
            if builder.count() == 0:
                valid_comment_found = False

        # Try to delete non-existing Comment - using delete_many
        # delete_many should return an empty list
        deleted_entities = self.backend.comments.delete_many(filters={'id': id_})
        self.assertEqual(
            deleted_entities, [],
            msg='No entities should have been deleted, since Comment id {} does not exist'.format(id_)
        )

        # Try to delete non-existing Comment - using delete
        # NotExistent should be raised, since no entities are found
        with self.assertRaises(exceptions.NotExistent) as exc:
            self.backend.comments.delete(comment_id=id_)
        self.assertIn("Comment with id '{}' not found".format(id_), str(exc.exception))

        # Try to delete existing and non-existing Comment - using delete_many
        # delete_many should return a list that *only* includes the existing Comment
        filters = {'id': {'in': [id_, comment_id]}}
        deleted_entities = self.backend.comments.delete_many(filters=filters)
        self.assertEqual([comment_id],
                         deleted_entities,
                         msg='Only Comment id {} should be returned from delete_many'.format(comment_id))

        # Make sure the existing Comment was deleted
        builder = orm.QueryBuilder().append(orm.Comment, filters={'uuid': comment_uuid})
        self.assertEqual(builder.count(), 0)

        # Get a non-existent Node
        valid_node_found = True
        id_ = 0
        while valid_node_found:
            id_ += 1
            builder = orm.QueryBuilder().append(orm.Node, filters={'id': id_})
            if builder.count() == 0:
                valid_node_found = False

        # Try to delete Comments filtering on non-existing dbnode_id
        # NotExistent should NOT be raised nor should any Comments be deleted
        comment_count_before = orm.QueryBuilder().append(orm.Comment).count()
        filters = {'dbnode_id': id_}
        self.backend.comments.delete_many(filters=filters)
        comment_count_after = orm.QueryBuilder().append(orm.Comment).count()
        self.assertEqual(
            comment_count_after,
            comment_count_before,
            msg='The number of comments changed after performing `delete_many`, '
            "while filtering for a non-existing 'dbnode_id'"
        )

    def test_delete_many_same_twice(self):
        """Test no exception is raised when entity is filtered by both `id` and `uuid`"""
        # Create comment
        comment = self.create_comment()
        comment.store()
        comment_id = comment.id
        comment_uuid = comment.uuid

        # Try to delete Comment by specifying both `id` and `uuid` for it - nothing should be raised
        self.backend.comments.delete_many(filters={'id': comment_id, 'uuid': comment_uuid})

        # Make sure comment is removed
        builder = orm.QueryBuilder().append(orm.Comment, filters={'uuid': comment_uuid})
        self.assertEqual(builder.count(), 0)

    def test_delete_wrong_type(self):
        """Test TypeError is raised when `filters` is wrong type"""
        with self.assertRaises(TypeError):
            self.backend.comments.delete(comment_id=None)

    def test_delete_many_wrong_type(self):
        """Test TypeError is raised when `filters` is wrong type"""
        with self.assertRaises(TypeError):
            self.backend.comments.delete_many(filters=None)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Delete trajectory symbols array from the repository and the reference in the attributes

Revision ID: ce56d84bcc35
Revises: 12536798d4d3
Create Date: 2019-01-21 15:35:07.280805

"""
# pylint: disable=invalid-name
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-member,no-name-in-module,import-error

import numpy

from alembic import op
from sqlalchemy.sql import table, column, select, func, text
from sqlalchemy import String, Integer, cast
from sqlalchemy.dialects.postgresql import UUID, JSONB

from aiida.backends.general.migrations import utils

# revision identifiers, used by Alembic.
revision = 'ce56d84bcc35'
down_revision = '12536798d4d3'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    # yapf:disable
    connection = op.get_bind()

    DbNode = table('db_dbnode', column('id', Integer), column('uuid', UUID), column('type', String),
                   column('attributes', JSONB))

    nodes = connection.execute(
        select([DbNode.c.id, DbNode.c.uuid]).where(
            DbNode.c.type == op.inline_literal('node.data.array.trajectory.TrajectoryData.'))).fetchall()

    for pk, uuid in nodes:
        connection.execute(
            text("""UPDATE db_dbnode SET attributes = attributes #- '{{array|symbols}}' WHERE id = {}""".format(pk)))
        utils.delete_numpy_array_from_repository(uuid, 'symbols')


def downgrade():
    """Migrations for the downgrade."""
    # yapf:disable
    connection = op.get_bind()

    DbNode = table('db_dbnode', column('id', Integer), column('uuid', UUID), column('type', String),
                   column('attributes', JSONB))

    nodes = connection.execute(
        select([DbNode.c.id, DbNode.c.uuid]).where(
            DbNode.c.type == op.inline_literal('node.data.array.trajectory.TrajectoryData.'))).fetchall()

    for pk, uuid in nodes:
        attributes = connection.execute(select([DbNode.c.attributes]).where(DbNode.c.id == pk)).fetchone()
        symbols = numpy.array(attributes['symbols'])
        utils.store_numpy_array_in_repository(uuid, 'symbols', symbols)
        key = op.inline_literal('{"array|symbols"}')
        connection.execute(DbNode.update().where(DbNode.c.id == pk).values(
            attributes=func.jsonb_set(DbNode.c.attributes, key, cast(list(symbols.shape), JSONB))))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration export files from old export versions to the newest, used by `verdi export migrate` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.cmdline.utils import echo
from aiida.tools.importexport.common.exceptions import DanglingLinkError

from .utils import verify_metadata_version, update_metadata
from .v01_to_v02 import migrate_v1_to_v2
from .v02_to_v03 import migrate_v2_to_v3
from .v03_to_v04 import migrate_v3_to_v4
from .v04_to_v05 import migrate_v4_to_v5
from .v05_to_v06 import migrate_v5_to_v6
from .v06_to_v07 import migrate_v6_to_v7

__all__ = ('migrate_recursively', 'verify_metadata_version')

MIGRATE_FUNCTIONS = {
    '0.1': migrate_v1_to_v2,
    '0.2': migrate_v2_to_v3,
    '0.3': migrate_v3_to_v4,
    '0.4': migrate_v4_to_v5,
    '0.5': migrate_v5_to_v6,
    '0.6': migrate_v6_to_v7
}


def migrate_recursively(metadata, data, folder):
    """
    Recursive migration of export files from v0.1 to newest version,
    See specific migration functions for detailed descriptions.

    :param metadata: the content of an export archive metadata.json file
    :param data: the content of an export archive data.json file
    :param folder: SandboxFolder in which the archive has been unpacked (workdir)
    """
    from aiida.tools.importexport import EXPORT_VERSION as newest_version

    old_version = verify_metadata_version(metadata)

    try:
        if old_version == newest_version:
            echo.echo_critical('Your export file is already at the newest export version {}'.format(newest_version))
        elif old_version in MIGRATE_FUNCTIONS:
            MIGRATE_FUNCTIONS[old_version](metadata, data, folder)
        else:
            echo.echo_critical('Cannot migrate from version {}'.format(old_version))
    except ValueError as exception:
        echo.echo_critical(exception)
    except DanglingLinkError:
        echo.echo_critical('Export file is invalid because it contains dangling links')

    new_version = verify_metadata_version(metadata)

    if new_version < newest_version:
        new_version = migrate_recursively(metadata, data, folder)

    return new_version



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Define the current configuration version and migrations."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__all__ = ('CURRENT_CONFIG_VERSION', 'OLDEST_COMPATIBLE_CONFIG_VERSION')

# The expected version of the configuration file and the oldest backwards compatible configuration version.
# If the configuration file format is changed, the current version number should be upped and a migration added.
# When the configuration file format is changed in a backwards-incompatible way, the oldest compatible version should
# be set to the new current version.
CURRENT_CONFIG_VERSION = 3
OLDEST_COMPATIBLE_CONFIG_VERSION = 3


class ConfigMigration(object):
    """Defines a config migration."""

    def __init__(self, migrate_function, version, version_oldest_compatible):
        """Construct a ConfigMigration

        :param migrate_function: function which migrates the configuration dictionary
        :param version: configuration version after the migration.
        :param version_oldest_compatible: oldest compatible configuration version after the migration.
        """
        self.migrate_function = migrate_function
        self.version = int(version)
        self.version_oldest_compatible = int(version_oldest_compatible)

    def apply(self, config):
        """Apply the migration to the configuration."""
        config = self.migrate_function(config)
        config.setdefault('CONFIG_VERSION', {})['CURRENT'] = self.version
        config.setdefault('CONFIG_VERSION', {})['OLDEST_COMPATIBLE'] = self.version_oldest_compatible
        return config


def _1_add_profile_uuid(config):
    """
    This adds the required values for a new default profile

        * PROFILE_UUID

    The profile uuid will be used as a general purpose identifier for the profile, in
    for example the RabbitMQ message queues and exchanges.
    """
    for profile in config.get('profiles', {}).values():
        from uuid import uuid4
        profile['PROFILE_UUID'] = uuid4().hex

    return config


def _2_simplify_default_profiles(config):
    """
    The concept of a different 'process' for a profile has been removed and as such the
    default profiles key in the configuration no longer needs a value per process ('verdi', 'daemon')
    We remove the dictionary 'default_profiles' and replace it with a simple value 'default_profile'
    """
    from aiida.manage.configuration import PROFILE

    default_profiles = config.pop('default_profiles', None)

    if default_profiles:
        config['default_profile'] = default_profiles['daemon']
    elif PROFILE is not None:
        config['default_profile'] = PROFILE.name

    return config


# Maps the initial config version to the ConfigMigration which updates it.
_MIGRATION_LOOKUP = {
    0: ConfigMigration(migrate_function=lambda x: x, version=1, version_oldest_compatible=0),
    1: ConfigMigration(migrate_function=_1_add_profile_uuid, version=2, version_oldest_compatible=0),
    2: ConfigMigration(migrate_function=_2_simplify_default_profiles, version=3, version_oldest_compatible=3)
}



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SQLA implementations for the Comment entity and collection."""
# pylint: disable=import-error,no-name-in-module
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from datetime import datetime
from sqlalchemy.orm.exc import NoResultFound

from aiida.backends.sqlalchemy import get_scoped_session
from aiida.backends.sqlalchemy.models import comment as models
from aiida.common import exceptions
from aiida.common import lang

from ..comments import BackendComment, BackendCommentCollection
from . import entities
from . import users
from . import utils


class SqlaComment(entities.SqlaModelEntity[models.DbComment], BackendComment):
    """Comment implementation for Sqla."""

    MODEL_CLASS = models.DbComment

    # pylint: disable=too-many-arguments
    def __init__(self, backend, node, user, content=None, ctime=None, mtime=None):
        """
        Construct a SqlaComment.

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :param ctime: The creation time as datetime object
        :param mtime: The modification time as datetime object
        :return: a Comment object associated to the given node and user
        """
        super(SqlaComment, self).__init__(backend)
        lang.type_check(user, users.SqlaUser)  # pylint: disable=no-member

        arguments = {
            'dbnode': node.dbmodel,
            'user': user.dbmodel,
            'content': content,
        }

        if ctime:
            lang.type_check(ctime, datetime, 'the given ctime is of type {}'.format(type(ctime)))
            arguments['ctime'] = ctime

        if mtime:
            lang.type_check(mtime, datetime, 'the given mtime is of type {}'.format(type(mtime)))
            arguments['mtime'] = mtime

        self._dbmodel = utils.ModelWrapper(models.DbComment(**arguments))

    def store(self):
        """Can only store if both the node and user are stored as well."""
        if self._dbmodel.dbnode.id is None or self._dbmodel.user.id is None:
            self._dbmodel.dbnode = None
            raise exceptions.ModificationNotAllowed('The corresponding node and/or user are not stored')

        super(SqlaComment, self).store()

    @property
    def ctime(self):
        return self._dbmodel.ctime

    @property
    def mtime(self):
        return self._dbmodel.mtime

    def set_mtime(self, value):
        self._dbmodel.mtime = value

    @property
    def node(self):
        return self.backend.nodes.from_dbmodel(self.dbmodel.dbnode)

    @property
    def user(self):
        return self.backend.users.from_dbmodel(self.dbmodel.user)

    def set_user(self, value):
        self._dbmodel.user = value

    @property
    def content(self):
        return self._dbmodel.content

    def set_content(self, value):
        self._dbmodel.content = value


class SqlaCommentCollection(BackendCommentCollection):
    """SqlAlchemy implementation for the CommentCollection."""

    ENTITY_CLASS = SqlaComment

    def create(self, node, user, content=None, **kwargs):
        """
        Create a Comment for a given node and user

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :return: a Comment object associated to the given node and user
        """
        return SqlaComment(self.backend, node, user, content, **kwargs)

    def delete(self, comment_id):
        """
        Remove a Comment from the collection with the given id

        :param comment_id: the id of the comment to delete
        :type comment_id: int

        :raises TypeError: if ``comment_id`` is not an `int`
        :raises `~aiida.common.exceptions.NotExistent`: if Comment with ID ``comment_id`` is not found
        """
        if not isinstance(comment_id, int):
            raise TypeError('comment_id must be an int')

        session = get_scoped_session()

        try:
            session.query(models.DbComment).filter_by(id=comment_id).one().delete()
            session.commit()
        except NoResultFound:
            session.rollback()
            raise exceptions.NotExistent("Comment with id '{}' not found".format(comment_id))

    def delete_all(self):
        """
        Delete all Comment entries.

        :raises `~aiida.common.exceptions.IntegrityError`: if all Comments could not be deleted
        """
        session = get_scoped_session()

        try:
            session.query(models.DbComment).delete()
            session.commit()
        except Exception as exc:
            session.rollback()
            raise exceptions.IntegrityError('Could not delete all Comments. Full exception: {}'.format(exc))

    def delete_many(self, filters):
        """
        Delete Comments based on ``filters``

        :param filters: similar to QueryBuilder filter
        :type filters: dict

        :return: (former) ``PK`` s of deleted Comments
        :rtype: list

        :raises TypeError: if ``filters`` is not a `dict`
        :raises `~aiida.common.exceptions.ValidationError`: if ``filters`` is empty
        """
        from aiida.orm import Comment, QueryBuilder

        # Checks
        if not isinstance(filters, dict):
            raise TypeError('filters must be a dictionary')
        if not filters:
            raise exceptions.ValidationError('filters must not be empty')

        # Apply filter and delete found entities
        builder = QueryBuilder().append(Comment, filters=filters, project='id').all()
        entities_to_delete = [_[0] for _ in builder]
        for entity in entities_to_delete:
            self.delete(entity)

        # Return list of deleted entities' (former) PKs for checking
        return entities_to_delete



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
""" Utility functions for import of AiiDA entities using SQLAlchemy backend """
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from uuid import UUID


def validate_uuid(given_uuid):
    """
    A simple check for the UUID validity.
    """
    try:
        parsed_uuid = UUID(given_uuid, version=4)
    except ValueError:
        # If not a valid UUID
        return False

    # Check if there was any kind of conversion of the hex during
    # the validation
    return str(parsed_uuid) == given_uuid
